"""
Test functionality for gval/homogenize/spatial_alignment.py
"""

# __all__ = ['*']
__author__ = "Fernando Aristizabal"

from pytest import raises
from pytest_cases import parametrize_with_cases
import xarray as xr
import numpy as np
import geopandas as gpd
from geopandas.testing import assert_geodataframe_equal

from gval.homogenize.spatial_alignment import (
    _matching_crs,
    _matching_spatial_indices,
    _rasters_intersect,
    _align_rasters,
    _spatial_alignment,
)
from gval.homogenize.numeric_alignment import _align_numeric_data_type
from gval.homogenize.rasterize import _rasterize_data
from gval.homogenize.vectorize import _vectorize_data
from gval.utils.exceptions import RasterMisalignment, RastersDontIntersect


@parametrize_with_cases(
    "candidate_map, benchmark_map, expected_match", glob="matching_crs"
)
def test_matching_crs(candidate_map, benchmark_map, expected_match):
    """Tests if two maps have matching CRSs"""
    matching = _matching_crs(candidate_map, benchmark_map)

    assert matching == expected_match, "CRSs of maps are not matching"


@parametrize_with_cases(
    "candidate_map, benchmark_map, expected_spatial_indices_matches",
    glob="matching_spatial_indices_success",
)
def test_matching_spatial_indices_success(
    candidate_map, benchmark_map, expected_spatial_indices_matches
):
    """Tests for matching indices in two xarrays"""
    matching = _matching_spatial_indices(candidate_map, benchmark_map)
    assert matching == expected_spatial_indices_matches, "Indices don't match expected"


@parametrize_with_cases(
    "candidate_map, benchmark_map, expected_spatial_indices_matches",
    glob="matching_spatial_indices_fail",
)
def test_matching_spatial_indices_fail(
    candidate_map, benchmark_map, expected_spatial_indices_matches
):
    """Tests for matching indices in two xarrays"""
    with raises(RasterMisalignment):
        _matching_spatial_indices(candidate_map, benchmark_map, raise_exception=True)


@parametrize_with_cases(
    "candidate_map, benchmark_map, expected_intersect",
    glob="rasters_intersect_no_exception",
)
def test_rasters_intersect_no_exception(
    candidate_map, benchmark_map, expected_intersect
):
    """Tests the intersection of rasters"""
    intersect = _rasters_intersect(candidate_map, benchmark_map)
    assert intersect == expected_intersect


@parametrize_with_cases(
    "candidate_map, benchmark_map, expected_intersect",
    glob="rasters_intersect_exception",
)
def test_rasters_intersect_exception(candidate_map, benchmark_map, expected_intersect):
    """Tests the intersection of rasters"""
    with raises(RastersDontIntersect):
        _rasters_intersect(candidate_map, benchmark_map, raise_exception=True)


@parametrize_with_cases(
    "candidate_map, benchmark_map, target_map, kwargs", glob="align_rasters"
)
def test_align_rasters(candidate_map, benchmark_map, target_map, kwargs):
    """Tests the alignment of rasters"""

    # This might raise value errors associated with
    cam, bem = _align_rasters(candidate_map, benchmark_map, target_map, **kwargs)

    # this tests for matching spatial indices
    _matching_spatial_indices(cam, bem, raise_exception=True)


@parametrize_with_cases(
    "candidate_map, benchmark_map, target_map, kwargs", glob="align_rasters_fail"
)
def test_align_rasters_fail(candidate_map, benchmark_map, target_map, kwargs):
    """Tests the alignment of rasters"""

    with raises(ValueError):
        _, _ = _align_rasters(candidate_map, benchmark_map, target_map, **kwargs)


@parametrize_with_cases(
    "candidate_map, benchmark_map, resampling, target_map",
    glob="align_rasters_fail_nodata",
)
def test_align_rasters_fail_nodata(
    candidate_map, benchmark_map, resampling, target_map
):
    """Tests the alignment of rasters"""

    with raises(ValueError):
        _, _ = _align_rasters(candidate_map, benchmark_map, target_map, resampling)


@parametrize_with_cases(
    "candidate_map, benchmark_map, target_map, kwargs",
    glob="spatial_alignment",
)
def test_spatial_alignment(candidate_map, benchmark_map, target_map, kwargs):
    """Tests spatial_alignment function"""

    cam, bem = _spatial_alignment(candidate_map, benchmark_map, target_map, **kwargs)

    try:
        # xr.align raises a value error if coordinates don't align
        xr.align(cam, bem, join="exact")
    except ValueError:
        assert False, "Candidate and benchmark failed to align"


@parametrize_with_cases(
    "candidate_map, benchmark_map, target_map, kwargs",
    glob="spatial_alignment_fail",
)
def test_spatial_alignment_fail(candidate_map, benchmark_map, target_map, kwargs):
    """Tests spatial_alignment function"""

    with raises(RastersDontIntersect):
        _, _ = _spatial_alignment(candidate_map, benchmark_map, target_map, **kwargs)


@parametrize_with_cases(
    "candidate_map, benchmark_map, rasterize_attributes, expected",
    glob="rasterize_vector_success",
)
def test_rasterize_vector_success(
    candidate_map, benchmark_map, rasterize_attributes, expected
):
    """Test rasterize vector success"""

    benchmark_raster = _rasterize_data(
        candidate_map=candidate_map,
        benchmark_map=benchmark_map,
        rasterize_attributes=rasterize_attributes,
    )

    if isinstance(benchmark_raster, xr.Dataset):
        assert benchmark_raster.band_1.shape == candidate_map.band_1.shape
    else:
        assert benchmark_raster.shape == candidate_map.shape

    xr.testing.assert_equal(benchmark_raster, expected)


@parametrize_with_cases(
    "candidate_map, benchmark_map, rasterize_attributes",
    glob="rasterize_vector_fail",
)
def test_rasterize_vector_fail(candidate_map, benchmark_map, rasterize_attributes):
    """Tests rasterize vector fail"""

    with raises(KeyError):
        _, _ = _rasterize_data(
            candidate_map=candidate_map,
            benchmark_map=benchmark_map,
            rasterize_attributes=rasterize_attributes,
        )


@parametrize_with_cases(
    "raster_map, expected",
    glob="vectorize_raster_success",
)
def test_vectorize_raster_success(raster_map, expected):
    """Test vectorize raster success"""

    vector_df = _vectorize_data(raster_data=raster_map)

    assert isinstance(vector_df, gpd.GeoDataFrame)
    assert_geodataframe_equal(
        vector_df.sort_values("geometry", ignore_index=True),
        expected.sort_values("geometry", ignore_index=True),
        check_index_type=False,
        check_dtype=False,
    )


@parametrize_with_cases(
    "raster_map",
    glob="vectorize_raster_fail",
)
def test_vectorize_raster_fail(raster_map):
    """Tests vectorize raster fail"""

    with raises(AttributeError):
        _ = _vectorize_data(raster_map)


@parametrize_with_cases(
    "candidate_map, benchmark_map, expected", glob="numeric_align_dataarrays"
)
def test_numeric_align_dataarrays(candidate_map, benchmark_map, expected):
    """Tests numeric alignment"""

    c, b = _align_numeric_data_type(candidate_map, benchmark_map)

    assert c.data.dtype == expected
    assert b.data.dtype == expected

    assert np.isclose(
        float(
            np.mean(
                xr.where(candidate_map == candidate_map.rio.nodata, 0, candidate_map)
            )
        ),
        float(np.mean(xr.where(c == c.rio.nodata, 0, c))),
    )
    assert np.isclose(
        float(
            np.mean(
                xr.where(benchmark_map == benchmark_map.rio.nodata, 0, benchmark_map)
            )
        ),
        float(np.mean(xr.where(b == b.rio.nodata, 0, b))),
    )


@parametrize_with_cases(
    "candidate_map, benchmark_map, expected", glob="numeric_align_datasets"
)
def test_numeric_align_datasets(candidate_map, benchmark_map, expected):
    """Tests numeric alignment for datasets"""

    c, b = _align_numeric_data_type(candidate_map, benchmark_map)

    for c_var, cand_var, b_var, bench_var in zip(
        c.data_vars, candidate_map.data_vars, b.data_vars, benchmark_map.data_vars
    ):
        assert c[c_var].data.dtype == expected
        assert b[b_var].data.dtype == expected

        assert np.isclose(
            float(
                np.mean(
                    xr.where(
                        candidate_map[cand_var] == candidate_map[cand_var].rio.nodata,
                        0,
                        candidate_map[cand_var],
                    )
                )
            ),
            float(np.mean(xr.where(c[c_var] == c[c_var].rio.nodata, 0, c[c_var]))),
        )

        assert np.isclose(
            float(
                np.mean(
                    xr.where(
                        benchmark_map[bench_var] == benchmark_map[bench_var].rio.nodata,
                        0,
                        benchmark_map[bench_var],
                    )
                )
            ),
            float(np.mean(xr.where(b[b_var] == b[b_var].rio.nodata, 0, b[b_var]))),
        )



# spatial_alignment.py
"""
Functions to check for and ensure spatial alignment of two xarray DataArrays

# To-Do
- make functions that check indices
"""

# __all__ = ['*']
__author__ = "Fernando Aristizabal"

from typing import Optional, Tuple, Union

import xarray as xr
from shapely.geometry import box
from rasterio.enums import Resampling
from odc.geo.xr import ODCExtensionDa

from gval.utils.loading_datasets import _handle_xarray_memory
from gval.utils.exceptions import RasterMisalignment, RastersDontIntersect

ODCExtensionDa


def _matching_crs(
    candidate_map: Union[xr.DataArray, xr.Dataset],
    benchmark_map: Union[xr.DataArray, xr.Dataset],
) -> bool:
    """
    Checks candidate and benchmark maps for matching crs's.

    Parameters
    ----------
    candidate_map : Union[xr.DataArray, xr.Dataset]
        Candidate map.
    benchmark_map : Union[xr.DataArray, xr.Dataset]
        Benchmark map.

    Returns
    -------
    bool
        Matching CRS's or not.
    """
    match = candidate_map.rio.crs == benchmark_map.rio.crs
    return match


def _matching_spatial_indices(
    candidate_map: Union[xr.DataArray, xr.Dataset],
    benchmark_map: Union[xr.DataArray, xr.Dataset],
    raise_exception: bool = False,
) -> bool:
    """
    Checks for matching indices in candidate and benchmark maps.

    TODO: How does this do with 3d DataArrays and Datasets?

    Parameters
    ----------
    candidate_map : Union[xr.DataArray, xr.Dataset]
        Candidate map.
    benchmark_map : Union[xr.DataArray, xr.Dataset]
        Benchmark map.

    Returns
    -------
    bool
        Whether or not maps are perfectly aligned.

    Raises
    ------
    RasterMisalignment
    """
    try:
        xr.align(candidate_map, benchmark_map, join="exact")
    except ValueError:
        aligned = False
    else:
        aligned = True

    # raises exception if not aligned and desired to raise
    if (not aligned) & (raise_exception):
        raise RasterMisalignment

    return aligned


def _rasters_intersect(
    candidate_map: Union[xr.DataArray, xr.Dataset],
    benchmark_map: Union[xr.DataArray, xr.Dataset],
    raise_exception: bool = False,
) -> bool:
    """
    Predicates spatial intersection of two `xr.DataArrays` or `xr.Datasets`.

    CRS's don't have to be the same.

    Parameters
    ----------
    candidate_map : Union[xr.DataArray, xr.Dataset]
        Candidate map.
    benchmark_map : Union[xr.DataArray, xr.Dataset]
        Benchmark map.

    Returns
    -------
    bool
        Tests spatial intersection.

    Raises
    ------
    RasterDontIntersect
        Rasters do not intersect.
    """

    # already matching crs's
    if _matching_crs(candidate_map, benchmark_map):
        candidate_map_bounds = candidate_map.rio.bounds()
        benchmark_map_bounds = benchmark_map.rio.bounds()

    # else transform bounds
    else:
        # use benchmark crs just to test for intersection
        target_crs = benchmark_map.rio.crs

        candidate_map_bounds = candidate_map.rio.transform_bounds(target_crs)
        benchmark_map_bounds = benchmark_map.rio.transform_bounds(target_crs)

    # convert bounds to shapely boxes
    candidate_map_box = box(*candidate_map_bounds)
    benchmark_map_box = box(*benchmark_map_bounds)

    # intersection as bool
    rasters_intersect_bool = candidate_map_box.intersects(benchmark_map_box)

    if (not rasters_intersect_bool) & raise_exception:
        raise RastersDontIntersect
    else:
        return rasters_intersect_bool


def _reproject_map(
    original_map: Union[xr.DataArray, xr.Dataset],
    target_map: Union[xr.DataArray, xr.Dataset],
    resampling: str,
) -> Union[xr.DataArray, xr.Dataset]:
    """

    Parameters
    ----------
    original_map: Union[xr.DataArray, xr.Dataset]
        Map to be reprojected
    target_map: Union[xr.DataArray, xr.Dataset]
        Map to use for extent, resolution, and spatial reference
    resampling: str
        Method to resample changing resolutions

    Returns
    -------
    Union[xr.DataArray, xr.Dataset]
        Reprojected map
    """

    is_dst = isinstance(original_map, xr.Dataset)

    nodata = target_map["band_1"].rio.nodata if is_dst else target_map.rio.nodata
    reproj = original_map.odc.reproject(
        target_map.odc.geobox, tight=True, dst_nodata=nodata, resampling=resampling
    )

    # Coordinates need to be aligned
    reproj_coords = (
        reproj.rename({"longitude": "x", "latitude": "y"})
        if "longitude" in reproj.coords
        else reproj
    )

    del reproj

    #  Coordinates are virtually the same but 1e-8 or so is rounded differently
    final_reproj = reproj_coords.assign_coords(
        {"x": target_map.coords["x"], "y": target_map.coords["y"]}
    )
    del reproj_coords
    return final_reproj


def _align_rasters(
    candidate_map: Union[xr.DataArray, xr.Dataset],
    benchmark_map: Union[xr.DataArray, xr.Dataset],
    target_map: Optional[Union[xr.DataArray, xr.Dataset, str]],
    resampling: Optional[Resampling] = Resampling.nearest,
) -> Tuple[Union[xr.DataArray, xr.Dataset], Union[xr.DataArray, xr.Dataset]]:
    """
    Reprojects raster to match a target map.

    Parameters
    ----------
    candidate_map : Union[xr.DataArray, xr.Dataset]
        Candidate map. Nodata value should be set.
    benchmark_map : Union[xr.DataArray, xr.Dataset]
        Benchmark map. Nodata value should be set.
    target_map : Optional[Union[xr.DataArray, xr.Dataset, str]]
        Target map as string that can have values "candidate" or "benchmark". Otherwise pass an `xr.DataArray` or `xr.Dataset` to align maps to.
    resampling : rasterio.enums.Resampling
        See :func:`rasterio.warp.reproject` for more details.
    **kwargs : dict or keyword arguments
        Dictionary or keyword arguments to be passed `rio.xarray.reproject_match()`. DEPRECATED: This was found to conflict with rxr.rio.reproject_matchrk
        ().
    Returns
    -------
    Tuple[Union[xr.DataArray, xr.Dataset], Union[xr.DataArray, xr.Dataset]]
        Tuple with aligned candidate and benchmark map respectively.

    Raises
    ------
    ValueError
        target_map argument only accepts xr.DataArray, xr.Dataset, 'candidate', 'benchmark'.
    """

    def ensure_nodata_value_is_set(dataset_or_dataarray):
        """
        Ensure nodata values are set.
        NOTE: this prevents arbitrary large numbers from being introduced into the aligned map.
        """

        err_message = (
            "Both candidate and benchmark maps need to have nodata values set."
        )

        # dataarray case
        if isinstance(dataset_or_dataarray, xr.DataArray):
            if dataset_or_dataarray.rio.nodata is None:
                raise ValueError(err_message)

        # dataset case
        elif isinstance(dataset_or_dataarray, xr.Dataset):
            for var_name in dataset_or_dataarray.data_vars.keys():
                if dataset_or_dataarray[var_name].rio.nodata is None:
                    raise ValueError(err_message)

    # ensure both candidate and benchmarks have nodata values set
    ensure_nodata_value_is_set(candidate_map)
    ensure_nodata_value_is_set(benchmark_map)

    # already matching crs's and indices
    if _matching_crs(candidate_map, benchmark_map) & _matching_spatial_indices(
        candidate_map, benchmark_map
    ):
        return candidate_map, benchmark_map

    # align benchmark and candidate to target
    elif isinstance(target_map, (xr.DataArray, xr.Dataset)):
        candidate_map = _reproject_map(candidate_map, target_map, resampling)
        benchmark_map = _reproject_map(benchmark_map, target_map, resampling)

    # match candidate to benchmark
    elif target_map == "benchmark":
        candidate_map = _reproject_map(candidate_map, benchmark_map, resampling)

    # match benchmark to candidate
    elif target_map == "candidate":
        benchmark_map = _reproject_map(benchmark_map, candidate_map, resampling)

    else:
        raise ValueError(
            "Target_map argument only accepts xr.DataArray or xr.Dataset or a string with values of 'candidate' or 'benchmark'."
        )

    return candidate_map, benchmark_map


def _spatial_alignment(
    candidate_map: Union[xr.DataArray, xr.Dataset],
    benchmark_map: Union[xr.DataArray, xr.Dataset],
    target_map: Optional[Union[xr.DataArray, xr.Dataset, str]] = "benchmark",
    resampling: Optional[Resampling] = Resampling.nearest,
) -> Tuple[Union[xr.DataArray, xr.Dataset], Union[xr.DataArray, xr.Dataset]]:
    """
    Reproject :class:`xarray.Dataset` objects

    .. note:: Only 2D/3D arrays with dimensions 'x'/'y' are currently supported.
        Others are appended as is.
        Requires either a grid mapping variable with 'spatial_ref' or
        a 'crs' attribute to be set containing a valid CRS.
        If using a WKT (e.g. from spatiareference.org), make sure it is an OGC WKT.

    Parameters
    ----------
    candidate_map: Union[xr.DataArray, xr.Dataset]
        Candidate map in xarray DataArray format.
    benchmark_map: Union[xr.DataArray, xr.Dataset]
        Benchmark map in xarray DataArray format.
    target_map: Optional[Union[xr.DataArray, xr.Dataset, str]], default = "benchmark"
        xarray object to match candidates and benchmarks to or str with 'candidate' or 'benchmark' as accepted values.
    resampling : rasterio.enums.Resampling
        See :func:`rasterio.warp.reproject` for more details.
    **kwargs: dict, optional
        Additional keyword arguments to pass into :func:`rasterio.warp.reproject`.
        DEPRECATED: This was found to have issues with spatial alignment if certain arguments were passed.

    Returns
    --------
    Tuple[Union[xr.DataArray, xr.Dataset], Union[xr.DataArray, xr.Dataset]]
        Tuple with candidate and benchmark map respectively.

    Raises
    ------
    RasterDontIntersect
        Rasters do not intersect.

    References
    ----------
    .. [1] `xr.reproject_match <https://corteva.github.io/rioxarray/stable/rioxarray.html#rioxarray.raster_dataset.RasterDataset.reproject>`_
    .. [2] `xr.reproject <https://corteva.github.io/rioxarray/stable/rioxarray.html#rioxarray.raster_dataset.RasterDataset.reproject>`_
    .. [2] `rasterio.warp.reproject <https://rasterio.readthedocs.io/en/latest/api/rasterio.warp.html#rasterio.warp.reproject>`_

    """

    # check if rasters intersect at all
    _rasters_intersect(candidate_map, benchmark_map, raise_exception=True)

    # reproject maps to align
    cam, bem = _align_rasters(candidate_map, benchmark_map, target_map, resampling)

    return tuple(map(lambda x: _handle_xarray_memory(x, make_temp=True), [cam, bem]))



"""
Computes probabilistic metrics.
"""

__author__ = "Fernando Aristizabal"

from typing import Optional, Any, Union

import numpy as np
import pandas as pd
import pandera.pandas as pa
from pandera.typing import DataFrame
import xskillscore as xs
import xarray as xr
import warnings

from gval.utils.schemas import Prob_metrics_df

PROB_METRICS = {
    "brier_score": xs.brier_score,
    "crps_ensemble": xs.crps_ensemble,
    "crps_gaussian": xs.crps_gaussian,
    "crps_quadrature": xs.crps_quadrature,
    "discrimination": xs.discrimination,
    "rank_histogram": xs.rank_histogram,
    "reliability": xs.reliability,
    "roc": xs.roc,
    "rps": xs.rps,
    "threshold_brier_score": xs.threshold_brier_score,
}

# None values are either used for required arguments or for optional arguments that are None by default
DEFAULT_METRIC_KWARGS = {
    "brier_score": {
        "observations": None,
        "forecasts": None,
        "member_dim": "member",
        "fair": False,
        "dim": None,
        "weights": None,
        "keep_attrs": False,
    },
    "crps_ensemble": {
        "observations": None,
        "forecasts": None,
        "member_weights": None,
        "issorted": False,
        "member_dim": "member",
        "dim": None,
        "weights": None,
        "keep_attrs": False,
    },
    "crps_gaussian": {
        "observations": None,
        "mu": None,
        "sig": None,
        "dim": None,
        "weights": None,
        "keep_attrs": False,
    },
    "crps_quadrature": {
        "observations": None,
        "cdf_or_dist": None,
        "xmin": None,
        "xmax": None,
        "tol": 1e-6,
        "dim": None,
        "keep_attrs": False,
    },
    "discrimination": {
        "observations": None,
        "forecasts": None,
        "dim": None,
        "probability_bin_edges": np.array([0.0, 0.2, 0.4, 0.6, 0.8, 1.0]),
    },
    "rank_histogram": {
        "observations": None,
        "forecasts": None,
        "dim": None,
        "member_dim": "member",
    },
    "reliability": {
        "observations": None,
        "forecasts": None,
        "dim": None,
        "probability_bin_edges": np.array([0.0, 0.2, 0.4, 0.6, 0.8, 1.0]),
        "keep_attrs": False,
    },
    "roc": {
        "observations": None,
        "forecasts": None,
        "bin_edges": "continuous",
        "dim": None,
        "drop_intermediate": False,
        "return_results": "area",
    },
    "rps": {
        "observations": None,
        "forecasts": None,
        "category_edges": None,
        "dim": None,
        "fair": False,
        "weights": None,
        "keep_attrs": False,
        "member_dim": "member",
        "input_distributions": None,
    },
    "threshold_brier_score": {
        "observations": None,
        "forecasts": None,
        "threshold": None,
        "issorted": False,
        "member_dim": "member",
        "dim": None,
        "weights": None,
        "keep_attrs": False,
    },
}

REQUIRED_METRIC_KWARGS = {
    "brier_score": {"observations", "forecasts"},
    "crps_ensemble": {"observations", "forecasts"},
    "crps_gaussian": {"observations", "mu", "sig"},
    "crps_quadrature": {"observations", "cdf_or_dist"},
    "discrimination": {"observations", "forecasts"},
    "rank_histogram": {"observations", "forecasts"},
    "reliability": {"observations", "forecasts"},
    "roc": {"observations", "forecasts"},
    "rps": {"observations", "forecasts", "category_edges"},
    "threshold_brier_score": {"observations", "forecasts", "threshold"},
}


@pa.check_types
def _compute_probabilistic_metrics(
    candidate_map: Union[xr.DataArray, xr.Dataset],
    benchmark_map: Union[xr.DataArray, xr.Dataset],
    metric_kwargs: dict,
    return_on_error: Optional[Any] = None,
) -> DataFrame[Prob_metrics_df]:
    """
    Computes probabilistic metrics from candidate and benchmark maps.

    Parameters
    ----------
    candidate_map : xr.DataArray or xr.Dataset
        Candidate map.
    benchmark_map : xr.DataArray or xr.Dataset
        Benchmark map.
    metric_kwargs : dict
        Dictionary of keyword arguments to metric functions. Keys must be metrics. Values are keyword arguments to metric functions. Don't pass keys or values for 'observations' or 'forecasts' as these are handled internally with `benchmark_map` and `candidate_map`, respectively. Available keyword arguments by metric are available in `DEFAULT_METRIC_KWARGS`. If values are None or empty dictionary, default values in `DEFAULT_METRIC_KWARGS` are used.
    return_on_error : Optional[Any], default = None
        Value to return within metrics dataframe if an error occurs when computing a metric. If None, the metric is not computed and None is returned. If 'error', the raised error is returned.

    Returns
    -------
    DataFrame[Prob_metrics_df]
        Probabilistic metrics Pandas DataFrame with computed xarray's per metric and sample.

    Raises
    ------
    ValueError
        If keyword argument is required for metric but not passed.
        If keyword argument is not available for metric but passed.
        If metric is not available.

    Warns
    -----
    UserWarning
        Warns if a metric cannot be computed. `return_on_error` determines whether the metric is not computed and None is returned or if the raised error is returned.

    References
    ----------
    .. [1] `Scoring rule <https://en.wikipedia.org/wiki/Scoring_rule#Examples_of_proper_scoring_rules>`_
    .. [2] `Strictly Proper Scoring Rules, Prediction, and Estimation <https://sites.stat.washington.edu/raftery/Research/PDF/Gneiting2007jasa.pdf>`_
    .. [3] `Properscoring/properscoring <https://github.com/properscoring/properscoring>`_
    .. [4] `xskillscore/xskillscore <https://xskillscore.readthedocs.io/en/stable/api.html#probabilistic-metrics>`_
    .. [5] `What is a proper scoring rule? <https://statisticaloddsandends.wordpress.com/2021/03/27/what-is-a-proper-scoring-rule/>`_
    """

    # Input handling for metric_kwargs variable ##########################################

    # ensure metric_kwargs are in REQUIRED_METRIC_KWARGS and DEFAULT_METRIC_KWARGS
    # also handle cases where values in metric_kwargs are None
    for metric in metric_kwargs.keys():
        # metric is available
        if metric in DEFAULT_METRIC_KWARGS.keys():
            # handle cases where values in metric_kwargs are None or empty
            if (metric_kwargs[metric] is None) | (metric_kwargs[metric] == dict()):
                for kw, val in DEFAULT_METRIC_KWARGS[metric].items():
                    kw_required = kw in REQUIRED_METRIC_KWARGS[metric]
                    dont_ignore_kwarg = kw not in {"observations", "forecasts"}

                    # when kw is required and not to be ignored
                    if (kw_required) & (dont_ignore_kwarg):
                        raise ValueError(
                            f"Keyword argument '{kw}' required for metric {metric}."
                        )

                    # if kw is optional and not to be ignored
                    if (not kw_required) & (dont_ignore_kwarg):
                        if metric_kwargs[metric] is None:
                            metric_kwargs[metric] = dict()

                        metric_kwargs[metric][kw] = val

            # add in observations if needed
            if "observations" in DEFAULT_METRIC_KWARGS[metric]:
                metric_kwargs[metric]["observations"] = benchmark_map

            # add in forecasts if needed
            if "forecasts" in DEFAULT_METRIC_KWARGS[metric]:
                metric_kwargs[metric]["forecasts"] = candidate_map

            # make sure all required arguments are passed
            for kw in REQUIRED_METRIC_KWARGS[metric]:
                if kw not in metric_kwargs[metric].keys():
                    raise ValueError(
                        f"Keyword argument '{kw}' required for metric {metric}."
                    )

            # make sure no extra arguments are passed
            for kw in metric_kwargs[metric].keys():
                if kw not in DEFAULT_METRIC_KWARGS[metric].keys():
                    raise ValueError(
                        f"Keyword argument '{kw}' not available for metric {metric}."
                    )

        # metric not available
        else:
            raise ValueError(f"Metric {metric} not available.")

    # Compute metrics ####################################################################

    # compute metric and return metric_df
    metric_df = pd.DataFrame(index=[0], columns=metric_kwargs.keys())
    for metric, kwargs in metric_kwargs.items():
        # get metric function
        metric_func = PROB_METRICS[metric]

        # compute metric value
        try:
            metric_df.loc[0, metric] = metric_func(**kwargs)
        except Exception as e:
            warnings.warn(f"Could not compute metric '{metric}': {e}", UserWarning)

            if return_on_error == "error":
                metric_df.loc[0, metric] = e
            else:
                metric_df.loc[0, metric] = return_on_error

        # select band name
        try:
            band_name = kwargs["member_dim"]  # member_dim passed as band name
        except KeyError:
            try:
                band_name = DEFAULT_METRIC_KWARGS[metric][
                    "member_dim"
                ]  # use default member_dim as band name
            except KeyError:
                band_name = "1"  # no member_dim passed and no default member_dim, use "1" as band name

        # add band column
        metric_df.insert(0, "band", band_name)

    # convert metric_dict to Prob_metrics_df
    metric_df = Prob_metrics_df(metric_df)

    return metric_df



"""
DataFrame Schemas with Pandera.
"""

from typing import List, Optional, Union

__author__ = "Fernando Aristizabal"

import pandas as pd
import pandera.pandas as pa
from pandera.typing import Series, Index, Int64
from shapely import Geometry


class Xrspatial_crosstab_df(pa.DataFrameModel):  # pragma: no cover
    """Defines the schema for output of `xrspatial.zonal.crosstab()`"""

    zone: Series[float]
    # TODO: Set schema to include any number of columns.
    # columns: pa.Column(name=Any, dtype=float, nullable=True, checks=None)
    # columns: Series[float] = pa.Field(regex=r"^(0|[1-9][0-9]*)$")

    class Config:
        coerce = True
        strict = False


class columns_method(pa.DataFrameModel):  # pragma: no cover
    """Defines base data frame model with columns method"""

    @classmethod
    def columns(cls) -> List[str]:
        """Gives access to columns from DataFrameModel"""
        return list(super().to_schema().columns.keys())


class Sample_identifiers(columns_method):  # pragma: no cover
    """Crosstab DF schema"""

    band: Series[str]
    # this is where we extend these identifiers to map or catalog level identifiers.
    # NOTE: Sample identifiers are currently columns. Should they be indices instead???
    # Add subsample
    idx: Index[Int64]

    class Config:
        coerce = True
        strict = True


class Subsample_identifiers(columns_method):  # pragma: no cover
    """Crosstab DF schema"""

    subsample: Optional[Series[str]]
    # this is where we extend these identifiers to map or catalog level identifiers.
    # NOTE: Sample identifiers are currently columns. Should they be indices instead???
    # Add subsample
    idx: Index[Int64]

    class Config:
        coerce = True
        strict = True


class Crosstab_df(Sample_identifiers, Subsample_identifiers):  # pragma: no cover
    """Crosstab DF schema

    Inherits columns from `Sample_identifiers`
    """

    candidate_values: Series = pa.Field(nullable=True)
    benchmark_values: Series = pa.Field(nullable=True)
    agreement_values: Optional[Series[float]]
    counts: Series[float]

    class Config:
        coerce = True
        strict = True


class Pivoted_crosstab_df(pa.DataFrameModel):  # pragma: no cover
    """Pivoted Crosstab DF schema"""

    row_idx: Index[Int64] = pa.Index(name="candidate_values")
    col_idx: Series[str]

    class Config:
        coerce = True
        strict = True

    @pa.dataframe_check
    def column_index_name(cls, df: pd.DataFrame) -> Series[bool]:
        """Checks that column index name is 'benchmark_values'"""
        return df.columns.name == "benchmark_values"


class Conditions_df(Sample_identifiers):  # pragma: no cover
    """Cateogrical conditions df

    Inherits columns from `Sample_identifiers`
    """

    tp: Optional[Series[float]]
    tn: Optional[Series[float]] = pa.Field(nullable=True)
    fp: Optional[Series[float]]
    fn: Optional[Series[float]]

    class Config:
        coerce = True
        ordered = False  # this is TEMP. Should set id's to indices?


class Metrics_df(Conditions_df):  # pragma: no cover
    """Metrics DF schema

    Inherits columns from `Conditions_df`
    """

    class Config:
        coerce = True
        strict = False  # set to False, bc columns could include any number of metrics


class Prob_metrics_df(Sample_identifiers, Subsample_identifiers):  # pragma: no cover
    """Probabilistic metrics DF schema"""

    metrics: Optional[Series[object]]

    class Config:
        coerce = False
        strict = False  # set to False, bc columns could include any number of metrics


class AttributeTrackingDf(pa.DataFrameModel):  # pragma: no cover
    """
    Defines the schema for output of `_attribute_tracking_xarray()`
    The attributes could be of any datatype. For instance, if your attributes are of float type,
    you can use Series[float] instead of Series[object].
    """

    # This is a basic schema considering attributes as object datatype.
    # If your attributes are of specific datatypes, you can modify accordingly.
    # Replace 'attribute_1', 'attribute_2' with your attribute names

    attribute_1_candidate: Optional[Series[object]]
    attribute_2_candidate: Optional[Series[object]]
    attribute_1_benchmark: Optional[Series[object]]
    attribute_2_benchmark: Optional[Series[object]]

    class Config:
        coerce = True
        strict = False

    @pa.dataframe_check
    def validate_column_suffixes(
        cls, df: pd.DataFrame, candidate_suffix: str, benchmark_suffix: str
    ) -> Series[bool]:
        """
        Checks that each column name in the dataframe ends with either '_candidate' or '_benchmark'.
        """
        return all(
            name.endswith((candidate_suffix, benchmark_suffix)) for name in df.columns
        )


class SubsamplingDf(pa.DataFrameModel):  # pragma: no cover
    """
    Defines the schema for subsampling DataFrame`
    """

    subsample_id: Series[int]
    geometry: Series[Geometry]
    subsample_type: Series[str]
    weights: Optional[Union[float, int]]  # Possibly for the future for sample weighting

    class Config:
        coerce = True
        strict = False



"""
Test functionality for computing_continuous_metrics.py
"""

# __all__ = ['*']
__author__ = "Fernando Aristizabal"


import pandas as pd
from pytest_cases import parametrize_with_cases
from pytest import raises
import numpy as np

from gval import ContStats as con_stat
from gval.comparison.compute_continuous_metrics import _compute_continuous_metrics


@parametrize_with_cases(
    "candidate_map, benchmark_map, metrics, expected_df",
    glob="compute_continuous_metrics_success",
)
def test_compute_continuous_metrics_success(
    candidate_map, benchmark_map, metrics, expected_df
):
    """tests continuous metrics functions"""

    # compute continuous metrics
    metrics_df = _compute_continuous_metrics(
        agreement_map=candidate_map - benchmark_map,
        candidate_map=candidate_map,
        benchmark_map=benchmark_map,
        metrics=metrics,
    )

    pd.testing.assert_frame_equal(
        metrics_df, expected_df, check_dtype=False
    ), "Compute statistics did not return expected values"


@parametrize_with_cases(
    "names, error_map, exception", glob="compute_continuous_statistics_fail"
)
def test_compute_continuous_statistics_fail(names, error_map, exception):
    """tests compute statistics fail function"""

    args = {
        "agreement_map": error_map,
        "metrics": names,
        "candidate_map": None,
        "benchmark_map": None,
    }

    with np.errstate(divide="ignore"):
        with raises(exception):
            # NOTE: Removed bc this should be handled within process_statistics.
            # stat_names = cat_stat.available_functions() if names == "all" else names
            _compute_continuous_metrics(**args)


@parametrize_with_cases("args, func", glob="register_continuous_function")
def test_register_continuous_function(args, func):
    """tests register func function"""

    con_stat.register_function(**args)(func)


@parametrize_with_cases(
    "args, func, exception", glob="register_continuous_function_fail"
)
def test_register_continuous_function_fail(args, func, exception):
    """tests register func fail function"""

    with raises(exception):
        con_stat.register_function(**args)(func)


@parametrize_with_cases("names, args, cls", glob="register_class_continuous")
def test_register_class_continuous(names, args, cls):
    """tests register class continuous function"""

    con_stat.register_function_class(**args)(cls)

    if [name in con_stat.registered_functions for name in names] != [True] * len(names):
        assert False, "Unable to register all class functions"


@parametrize_with_cases("args, cls, exception", glob="register_class_continuous_fail")
def test_register_class_continuous_fail(args, cls, exception):
    """tests register class continuous fail function"""

    with raises(exception):
        con_stat.register_function_class(**args)(cls)


@parametrize_with_cases("name, params", glob="get_param_continuous")
def test_get_param_continuous(name, params):
    """tests get param continuous function"""

    _params = con_stat.get_parameters(name)
    assert _params == params


@parametrize_with_cases("name", glob="get_param_continuous_fail")
def test_get_param_continuous_fail(name):
    """tests get param continuous fail function"""

    with raises(KeyError):
        _ = con_stat.get_parameters(name)


def test_get_all_param_continuous():
    """tests get all params function"""

    try:
        con_stat.get_all_parameters()
    except KeyError:
        assert False, "Signature dict not present or keys changed"


def test_available_functions_continuous():
    """tests get available functions"""

    a_funcs = con_stat.available_functions()

    assert isinstance(a_funcs, list)



from typing import Union

import geopandas as gpd
import numpy as np
from shapely.geometry import shape
from rasterio.features import shapes
import xarray as xr


def _vectorize_data(
    raster_data: Union[xr.Dataset, xr.DataArray],
) -> gpd.GeoDataFrame:
    """

    Parameters
    ----------
    raster_data: Union[xr.Dataset, xr.DataArray]
        Raster map to convert to vectorized map

    Returns
    -------
    gpd.GeoDataFrame
        Vectorized data

    """

    # Allowed numerical datatypes in vectorize operation
    dtypes = [np.int16, np.int32, np.uint8, np.uint16, np.float32]

    def get_dtype(dtype):
        """
        Conform datatype to alloable vectorize dtype

        Parameters
        ----------
        dtype : type
            Numerical datatype of raster data

        Returns
        -------
        type
            Type to use for vectorizing
        """

        if dtype not in dtypes:
            if np.issubdtype(dtype, np.integer):
                dtype = np.int32
            else:
                dtype = np.float32

        return dtype

    # Get bands, nodata, and dtype for data array or every var from a dataset
    if isinstance(raster_data, xr.DataArray):
        if len(raster_data.shape) > 2:
            iter_bands = [
                f"band_{coord.values}" for coord in raster_data.coords["band"]
            ]
            indices = range(raster_data.shape[0])
        else:
            iter_bands, indices = ["band_1"], [slice(0, raster_data.shape[0])]
        nodata_vals = [raster_data.rio.nodata] * len(iter_bands)
        dtypes = [get_dtype(raster_data.dtype)] * len(iter_bands)

    else:
        iter_bands = indices = [d_var for d_var in raster_data.data_vars]
        nodata_vals = [raster_data[d_var].rio.nodata for d_var in raster_data.data_vars]
        dtypes = [
            get_dtype(raster_data[d_var].dtype) for d_var in raster_data.data_vars
        ]

    bands, values, geometry = [], [], []

    # Iterate over values and vectorize shapes
    for idx, band, nodata, dtype in zip(indices, iter_bands, nodata_vals, dtypes):
        shps = shapes(
            raster_data[idx].astype(np.float32),
            transform=raster_data[idx].rio.transform(),
        )

        for shp in shps:
            if shp[1] != nodata and not np.isnan(shp[1]):
                values.append(shp[1])
                geometry.append(shape(shp[0]))
                bands.append(band.split("_")[-1])

    gdf_vec = gpd.GeoDataFrame({"band": bands, "values": values, "geometry": geometry})
    gdf_vec.crs = raster_data[idx].rio.crs

    del raster_data

    return gdf_vec



import numpy as np
from pytest_cases import parametrize_with_cases
from pytest import raises

from gval.statistics.categorical_statistics import CategoricalStatistics

cat_stat = CategoricalStatistics()


@parametrize_with_cases("names, args, funcs, expected", glob="categorical_statistics")
def test_categorical_statistics(names, args, funcs, expected):
    """tests categorical statistics functions"""

    np.testing.assert_equal(
        [
            func(*[args[p] for p in cat_stat.get_parameters(name)])
            for func, name in zip(funcs, names)
        ],
        expected,
    ), "Compute statistics did not return expected values"


@parametrize_with_cases("names, args, expected", glob="compute_statistics")
def test_compute_statistics(names, args, expected):
    """tests compute statistics fail function"""

    # NOTE: Removed bc this should be handled within process_statistics.
    # stat_names = cat_stat.available_functions() if names == "all" else names

    np.testing.assert_equal(
        cat_stat.process_statistics(names, **args)[
            0
        ],  # NOTE: Only testing metric value returns.
        expected,
    ), "Compute statistics did not return expected values"


@parametrize_with_cases("names, args, exception", glob="compute_statistics_fail")
def test_compute_statistics_fail(names, args, exception):
    """tests compute statistics fail function"""

    with np.errstate(divide="ignore"):
        with raises(exception):
            # NOTE: Removed bc this should be handled within process_statistics.
            # stat_names = cat_stat.available_functions() if names == "all" else names
            cat_stat.process_statistics(names, **args)


@parametrize_with_cases("args, func", glob="register_function")
def test_register_function(args, func):
    """tests register func function"""

    cat_stat.register_function(**args)(func)


@parametrize_with_cases("args, func, exception", glob="register_function_fail")
def test_register_function_fail(args, func, exception):
    """tests register func fail function"""

    with raises(exception):
        cat_stat.register_function(**args)(func)


@parametrize_with_cases("names, args, cls", glob="register_class")
def test_register_class(names, args, cls):
    """tests register class function"""

    cat_stat.register_function_class(**args)(cls)

    if [name in cat_stat.registered_functions for name in names] != [True] * len(names):
        assert False, "Unable to register all class functions"


@parametrize_with_cases("args, cls, exception", glob="register_class_fail")
def test_register_class_fail(args, cls, exception):
    """tests register class fail function"""

    with raises(exception):
        cat_stat.register_function_class(**args)(cls)


@parametrize_with_cases("name, params", glob="get_param")
def test_get_param(name, params):
    """tests get param function"""

    _params = cat_stat.get_parameters(name)
    assert _params == params


@parametrize_with_cases("name", glob="get_param_fail")
def test_get_param_fail(name):
    """tests get param fail function"""

    with raises(KeyError):
        _ = cat_stat.get_parameters(name)


def test_get_all_param():
    """tests get all params function"""

    try:
        cat_stat.get_all_parameters()
    except KeyError:
        assert False, "Signature dict not present or keys changed"


def test_available_functions():
    """tests get available functions"""

    a_funcs = cat_stat.available_functions()

    assert isinstance(a_funcs, list)



"""
Continuous Statistics Functions From Error Based Agreement Maps.
"""

from numbers import Number
from typing import Union

import xarray as xr
import numpy as np

from gval.statistics.continuous_stat_utils import convert_output


@convert_output
def mean_absolute_error(error: Union[xr.DataArray, xr.Dataset]) -> Number:
    """
    Compute mean absolute error (MAE).

    Either `error` or (`candidate_map` and `benchmark_map`) must be provided.

    Parameters
    ----------
    error : Union[xr.DataArray, xr.Dataset]
        Candidate minus benchmark error.

    Returns
    -------
    MAE : Number
        Mean absolute error.

    References
    ----------
    .. [1] `Mean absolute error <https://en.wikipedia.org/wiki/Mean_absolute_error>`_
    """
    return np.abs(error).mean()


@convert_output
def mean_squared_error(error: Union[xr.DataArray, xr.Dataset]) -> Number:
    """
    Compute mean squared error (MSE).

    Either `error` or (`candidate_map` and `benchmark_map`) must be provided.

    Parameters
    ----------
    error : Union[xr.DataArray, xr.Dataset]
        Candidate minus benchmark error.

    Returns
    -------
    MSE : Number
        Mean squared error.

    References
    ----------
    .. [1] `Mean squared error <https://en.wikipedia.org/wiki/Mean_squared_error>`_
    """

    return (error**2).mean()


@convert_output
def root_mean_squared_error(error: Union[xr.DataArray, xr.Dataset]) -> Number:
    """
    Compute root mean squared error (RMSE).

    Either `error` or (`candidate_map` and `benchmark_map`) must be provided.

    Parameters
    ----------
    error : Union[xr.DataArray, xr.Dataset]
        Candidate minus benchmark error.

    Returns
    -------
    RMSE : Number
        Root mean squared error.

    References
    ----------
    .. [1] `Root mean square deviation <https://en.wikipedia.org/wiki/Root-mean-square_deviation>`_
    """

    return np.sqrt((error**2).mean())


@convert_output
def mean_signed_error(error: Union[xr.DataArray, xr.Dataset]) -> Number:
    """
    Compute mean signed error (MSiE).

    Either `error` or (`candidate_map` and `benchmark_map`) must be provided.

    Parameters
    ----------
    error : Union[xr.DataArray, xr.Dataset]
        Candidate minus benchmark error.

    Returns
    -------
    MSiE : Number
        Mean signed error.

    References
    ----------
    .. [1] `Mean signed error <https://en.wikipedia.org/wiki/Mean_signed_difference>`_
    """

    return error.mean()


@convert_output
def mean_percentage_error(
    error: Union[xr.DataArray, xr.Dataset],
    benchmark_map: Union[xr.DataArray, xr.Dataset],
) -> Number:
    """
    Compute mean percentage error (MPE) as a fraction.

    Either (`error` and `benchmark_map`) or (`candidate_map` and `benchmark_map`) must be provided.

    Parameters
    ----------
    error : Union[xr.DataArray, xr.Dataset]
        Candidate minus benchmark error.
    benchmark_map : Union[xr.DataArray, xr.Dataset]
        Benchmark map.

    Returns
    -------
    MPE : Number
        Mean percentage error as a fraction.

    References
    ----------
    .. [1] `Mean percentage error <https://en.wikipedia.org/wiki/Mean_percentage_error>`_
    """
    return (error / benchmark_map.mean()).mean()


@convert_output
def mean_absolute_percentage_error(
    error: Union[xr.DataArray, xr.Dataset],
    benchmark_map: Union[xr.DataArray, xr.Dataset],
    epsilon: float = 1e-10,
) -> Number:
    """
    Compute mean absolute percentage error (MAPE) as a fraction.

    Either (`error` and `benchmark_map`) or (`candidate_map` and `benchmark_map`) must be provided.

    Parameters
    ----------
    error : Union[xr.DataArray, xr.Dataset]
        Candidate minus benchmark error.
    benchmark_map : Union[xr.DataArray, xr.Dataset]
        Benchmark map.
    epsilon : float, default = 1e-10
        Small value to avoid division by zero.

    Returns
    -------
    MAPE : Number
        Mean absolute percentage error as a fraction.

    References
    ----------
    .. [1] `Mean absolute percentage error <https://en.wikipedia.org/wiki/Mean_absolute_percentage_error>`_
    """
    mask = np.abs(benchmark_map) < epsilon
    benchmark_map = benchmark_map.where(~mask, other=epsilon)
    return np.abs(error / benchmark_map).mean()


@convert_output
def mean_normalized_root_mean_squared_error(
    error: Union[xr.DataArray, xr.Dataset],
    benchmark_map: Union[xr.DataArray, xr.Dataset],
) -> Number:
    """
    Compute mean normalized root mean squared error (NRMSE).

    Either (`error` and `benchmark_map`) or (`candidate_map` and `benchmark_map`) must be provided.

    Parameters
    ----------
    error : Union[xr.DataArray, xr.Dataset]
        Candidate minus benchmark error.
    benchmark_map : Union[xr.DataArray, xr.Dataset]
        Benchmark map.

    Returns
    -------
    mNRMSE : Number
        Mean normalized root mean squared error.

    References
    ----------
    .. [1] `Normalized root-mean-square deviation <https://en.wikipedia.org/wiki/Root-mean-square_deviation#Normalized_root-mean-square_deviation>`_
    """

    return np.sqrt((error**2).mean()) / benchmark_map.mean()


@convert_output
def range_normalized_root_mean_squared_error(
    error: Union[xr.DataArray, xr.Dataset],
    benchmark_map: Union[xr.DataArray, xr.Dataset],
) -> Number:
    """
    Compute range normalized root mean squared error (RNRMSE).

    Either (`error` and `benchmark_map`) or (`candidate_map` and `benchmark_map`) must be provided.

    Parameters
    ----------
    error : Union[xr.DataArray, xr.Dataset]
        Candidate minus benchmark error.
    benchmark_map : Union[xr.DataArray, xr.Dataset]
        Benchmark map.

    Returns
    -------
    rNRMSE : Number
        Range normalized root mean squared error.

    References
    ----------
    .. [1] `Normalized root-mean-square deviation <https://en.wikipedia.org/wiki/Root-mean-square_deviation#Normalized_root-mean-square_deviation>`_
    """
    return np.sqrt((error**2).mean()) / (benchmark_map.max() - benchmark_map.min())


@convert_output
def mean_normalized_mean_absolute_error(
    error: Union[xr.DataArray, xr.Dataset],
    benchmark_map: Union[xr.DataArray, xr.Dataset],
) -> Number:
    """
    Compute mean normalized mean absolute error (NMAE).

    Either (`error` and `benchmark_map`) or (`candidate_map` and `benchmark_map`) must be provided.

    Parameters
    ----------
    error : Union[xr.DataArray, xr.Dataset]
        Candidate minus benchmark error.
    benchmark_map : Union[xr.DataArray, xr.Dataset]
        Benchmark map.

    Returns
    -------
    NMAE : Number
        Normalized mean absolute error.

    References
    ----------
    .. [1] `Normalized mean absolute error <https://en.wikipedia.org/wiki/Mean_absolute_error#Normalized_mean_absolute_error>`_
    """
    return np.abs(error).mean() / benchmark_map.mean()


@convert_output
def range_normalized_mean_absolute_error(
    error: Union[xr.DataArray, xr.Dataset],
    benchmark_map: Union[xr.DataArray, xr.Dataset],
) -> Number:
    """
    Compute range normalized mean absolute error (RNMAE).

    Either (`error` and `benchmark_map`) or (`candidate_map` and `benchmark_map`) must be provided.

    Parameters
    ----------
    error : Union[xr.DataArray, xr.Dataset]
        Candidate minus benchmark error.
    benchmark_map : Union[xr.DataArray, xr.Dataset]
        Benchmark map.

    Returns
    -------
    rNMAE : Number
        Range normalized mean absolute error.

    References
    ----------
    .. [1] `Normalized mean absolute error <https://en.wikipedia.org/wiki/Mean_absolute_error#Normalized_mean_absolute_error>`_
    """
    return np.abs(error).mean() / (benchmark_map.max() - benchmark_map.min())


@convert_output
def coefficient_of_determination(
    error: Union[xr.DataArray, xr.Dataset],
    benchmark_map: Union[xr.DataArray, xr.Dataset],
) -> Number:
    """
    Compute coefficient of determination (R2).

    Either (error and benchmark_map) or (candidate_map and benchmark_map) must be provided.

    Parameters
    ----------
    error : Union[xr.DataArray, xr.Dataset]
        Candidate minus benchmark error.
    benchmark_map : Union[xr.DataArray, xr.Dataset]
        Benchmark map.

    Returns
    -------
    R2 : Number
        Coefficient of determination.

    References
    ----------
    .. [1] `Coefficient of determination <https://en.wikipedia.org/wiki/Coefficient_of_determination>`_
    """

    return 1 - (error**2).sum() / ((benchmark_map - benchmark_map.mean()) ** 2).sum()


@convert_output
def symmetric_mean_absolute_percentage_error(
    error: Union[xr.DataArray, xr.Dataset],
    candidate_map: Union[xr.DataArray, xr.Dataset],
    benchmark_map: Union[xr.DataArray, xr.Dataset],
) -> Number:
    """
    Compute symmetric mean absolute percentage error (sMAPE) as a fraction.

    Both `candidate_map` and `benchmark_map` must be provided. `error` can be provided to avoid recomputing it.

    Parameters
    ----------
    error : Union[xr.DataArray, xr.Dataset]
        Candidate minus benchmark error.
    candidate_map : Union[xr.DataArray, xr.Dataset]
        Candidate map.
    benchmark_map : Union[xr.DataArray, xr.Dataset]
        Benchmark map.

    Returns
    -------
    sMAPE : Number
        Symmetric mean absolute percentage error as a fraction.

    References
    ----------
    .. [1] `Symmetric mean absolute percentage error <https://en.wikipedia.org/wiki/Symmetric_mean_absolute_percentage_error>`_
    """
    return (
        2 * np.abs(error).sum() / (np.abs(candidate_map) + np.abs(benchmark_map)).sum()
    )



"""
Test functionality for gval/compare.py
"""

# __all__ = ['*']
__author__ = "Fernando Aristizabal"

import numpy as np
import pandas as pd
from pytest_cases import parametrize_with_cases
from pytest import raises

from gval.subsampling.subsampling import subsample


@parametrize_with_cases("df, args", glob="create_sampling_dataframes")
def test_create_sampling_dataframes(df, args):
    df_out = df.gval.create_subsampling_df(**args)
    args["inplace"] = True
    df.gval.create_subsampling_df(**args)

    assert np.sum(np.sum(df != df_out)) == 0

    if args["subsampling_weights"] is not None:
        assert "weights" in df.columns


@parametrize_with_cases("df, args", glob="create_sampling_dataframes_fail")
def test_create_sampling_dataframes_fail(df, args):
    df.crs = None
    df.drop(columns={"geometry"}, inplace=True)

    with raises(TypeError):
        _ = df.gval.create_subsampling_df(**args)


@parametrize_with_cases(
    "candidate, benchmark, subsample_df, subsample_type, expected_length, sample_percent",
    glob="subsampling",
)
def test_subsampling(
    candidate, benchmark, subsample_df, subsample_type, expected_length, sample_percent
):
    subsample_df.gval.create_subsampling_df(
        subsampling_type=subsample_type, inplace=True
    )

    results = subsample(
        candidate=candidate,
        benchmark=benchmark,
        subsampling_df=subsample_df,
    )

    assert len(results) == expected_length
    for res, perc in zip(results, sample_percent):
        res[0].attrs["sample_percentage"] == perc


@parametrize_with_cases(
    "candidate, benchmark, subsample_df, exception", glob="subsampling_fail"
)
def test_subsampling_fail(candidate, benchmark, subsample_df, exception):
    if exception == ValueError:
        subsample_df.crs = None
    else:
        subsample_df.gval.create_subsampling_df(
            subsampling_type="include", inplace=True
        )

    with raises(exception):
        _ = subsample(
            candidate=candidate, benchmark=benchmark, subsampling_df=subsample_df
        )


@parametrize_with_cases(
    "candidate, benchmark, subsample_df, expected_df, sampling_average",
    glob="categorical_subsample",
)
def test_categorical_subsampling(
    candidate, benchmark, subsample_df, expected_df, sampling_average
):
    subsample_df.gval.create_subsampling_df(subsampling_type="include", inplace=True)

    ag, ctab, met = candidate.gval.categorical_compare(
        benchmark_map=benchmark,
        positive_categories=[2],
        negative_categories=[0, 1],
        subsampling_df=subsample_df,
        subsampling_average=sampling_average,
    )

    pd.testing.assert_frame_equal(
        met, expected_df, check_dtype=False
    ), "Compute statistics did not return expected values"

    assert len(ag) == 2


@parametrize_with_cases(
    "candidate, benchmark, subsample_df, expected_df, sampling_average",
    glob="continuous_subsample",
)
def test_continuous_subsampling(
    candidate, benchmark, subsample_df, expected_df, sampling_average
):
    subsample_df.gval.create_subsampling_df(
        subsampling_type="exclude",
        subsampling_weights=[2, 1],
        inplace=True,
        crs="EPSG:4326",
    )

    ag, met = candidate.gval.continuous_compare(
        benchmark_map=benchmark,
        metrics=["mean_percentage_error"],
        subsampling_df=subsample_df,
        subsampling_average=sampling_average,
    )

    pd.testing.assert_frame_equal(
        met, expected_df, check_dtype=False
    ), "Compute statistics did not return expected values"

    assert len(ag) == 2



"""
Categorical Statistics Class
"""

from typing import Union, Tuple
from functools import wraps
import inspect

import numpy as np
from numba import vectorize

from gval.statistics.base_statistics import BaseStatistics
import gval.statistics.categorical_stat_funcs as cs


class CategoricalStatistics(BaseStatistics):
    """
    Class for Running Categorical Statistics on Agreement Maps

    Attributes
    ----------
    registered_functions : dict
        Available statistical functions with names as keys and parameters as values
    """

    def __init__(self):
        # Automatically populates and numba vectorizes all functions in categorical_stat_funcs.py

        self.required_param = 1
        self.optional_param = 0

        self._func_names = [
            fn
            for fn in dir(cs)
            if len(fn) > 5 and "__" not in fn and "Number" not in fn
        ]
        self._funcs = [getattr(cs, name) for name in self._func_names]

        for name, func in zip(self._func_names, self._funcs):
            setattr(self, name, vectorize(nopython=True)(func))

        self._signature_validation = {
            "names": {
                "tp": self.required_param,
                "tn": self.optional_param,
                "fp": self.required_param,
                "fn": self.required_param,
            },
            "required": [
                self.required_param,
                self.optional_param,
                self.required_param,
                self.required_param,
            ],
            "param_types": ["int", "float", "Number"],
            "return_type": [float],
            "no_of_args": [2, 3, 4],
        }

        self.registered_functions = {
            name: {"params": [param for param in inspect.signature(func).parameters]}
            for name, func in zip(self._func_names, self._funcs)
        }

    def available_functions(self) -> list:
        """
        Lists all available functions

        Returns
        -------
        List of available functions
        """
        return list(self.registered_functions.keys())

    def get_all_parameters(self):
        """
        Get all the possible arguments

        Returns
        -------
        List of all possible arguments for functions
        """

        return list(self._signature_validation["names"].keys())

    def register_function(self, name: str, vectorize_func: bool = False):
        """
        Register decorator function in statistics class

        Parameters
        ----------
        name: str
            Name of function to register in statistics class
        vectorize_func: bool
            Whether to vectorize the function

        Returns
        -------
        Decorator function
        """

        def decorator(func):
            self.function_signature_check(func)

            if name not in self.registered_functions:
                self.registered_functions[name] = {
                    "params": [
                        param
                        for param in inspect.signature(func).parameters
                        if param != "self"
                    ]
                }
                # vectorize function if vectorize_func is True
                r_func = (
                    vectorize(nopython=True)(func) if vectorize_func is True else func
                )
                setattr(self, name, r_func)
            else:
                raise KeyError("This function name already exists")

            @wraps(func)
            def wrapper(*args, **kwargs):  # pragma: no cover
                result = func(*args, **kwargs)

                return result

            return wrapper

        return decorator

    def register_function_class(self, vectorize_func: bool = False):
        """
        Register decorator function for an entire class

        Parameters
        ----------
        vectorize_func: bool
            Whether to vectorize the function

        """

        def decorator(dec_self: object):
            """
            Decorator for wrapper

            Parameters
            ----------
            dec_self: object
                Class to register stat functions
            """

            for name, func in inspect.getmembers(dec_self, inspect.isfunction):
                if name not in self.registered_functions:
                    self.function_signature_check(func)
                    self.registered_functions[name] = {
                        "params": [
                            param
                            for param in inspect.signature(func).parameters
                            if param != "self"
                        ]
                    }
                    # vectorize funciton if vectorize_func is True
                    r_func = (
                        vectorize(nopython=True)(func)
                        if vectorize_func is True
                        else func
                    )
                    setattr(self, name, r_func)
                else:
                    raise KeyError("This function name already exists")

        return decorator

    def function_signature_check(self, func):
        """
        Validates signature of registered function

        Parameters
        ----------
        func: function
            Function to check the signature of
        """
        signature = inspect.signature(func)
        names = self._signature_validation["names"]
        param_types = self._signature_validation["param_types"]
        return_type = self._signature_validation["return_type"]
        no_of_args = self._signature_validation["no_of_args"]

        # Checks if param names, type, and return type are in valid list
        # Considered no validation if either are empty
        for key, val in signature.parameters.items():
            if (key not in names and len(names) > 0) or (
                not str(val).split(": ")[-1].split(".")[-1] in param_types
                and len(param_types) > 0
            ):
                raise TypeError(
                    "Wrong parameters in function: \n"
                    f"Valid Names: {names} \n"
                    f"Valid Types: {param_types} \n"
                )

        if len(no_of_args) > 0 and len(signature.parameters) not in no_of_args:
            raise TypeError(
                "Wrong number of parameters: \n"
                f"Valid number of parameters: {no_of_args}"
            )

        if signature.return_annotation not in return_type and len(return_type) > 0:
            raise TypeError("Wrong return type \n" f"Valid return Type {return_type}")

    def get_parameters(self, func_name: str) -> list:
        """
        Get parameters of registered function

        Parameters
        ----------
        func_name: str


        Returns
        -------
        List of parameter names for the associated function
        """

        if func_name in self.registered_functions:
            return self.registered_functions[func_name]["params"]
        else:
            raise KeyError("Statistic not found in registered functions")

    def process_statistics(
        self, func_names: Union[str, list], **kwargs
    ) -> Tuple[float, str]:
        """

        Parameters
        ----------
        func_names: Union[str, list]
            Name of registered function to run
        **kwargs: dict or keyword arguments
            Dictionary or keyword arguments of to pass to metric functions.

        Returns
        -------
        Tuple[float, str]
            Tuple with metric values and metric names.
        """

        func_names = (
            list(self.registered_functions.keys())
            if func_names == "all"
            else func_names
        )
        func_list = [func_names] if isinstance(func_names, str) else func_names

        return_stats, return_funcs = [], []
        for name in func_list:
            if name in self.registered_functions:
                params = self.get_parameters(name)
                required = self._signature_validation["required"]

                func = getattr(self, name)

                # Necessary for numba functions which cannot accept keyword arguments
                func_args, skip_function, return_nan = [], False, False
                for param, req in zip(params, required):
                    if param in kwargs:
                        func_args.append(kwargs[param])
                    elif not self._signature_validation["names"][param]:
                        skip_function = True
                        break
                    else:
                        return_nan = True
                        break

                if skip_function:
                    continue

                with np.errstate(divide="ignore", invalid="ignore"):
                    stat_val = np.nan if return_nan else func(*func_args)

                return_stats.append(stat_val)
                return_funcs.append(name)

            else:
                raise KeyError(f"Statistic, {name}, not found in registered functions")

        return return_stats, return_funcs



"""
Test functionality for gval/homogenize/spatial_alignment.py
"""

# __all__ = ['*']
__author__ = "Fernando Aristizabal"

from pytest import raises
from pytest_cases import parametrize_with_cases
import pandas as pd

from gval.attributes.attributes import _attribute_tracking_xarray
from tests.conftest import _assert_pairing_dict_equal


@parametrize_with_cases(
    "candidate_map, benchmark_map, agreement_map, candidate_suffix, benchmark_suffix, candidate_include, candidate_exclude, benchmark_include, benchmark_exclude, expected_df, expected_attr",
    glob="attribute_tracking",
)
def test_attribute_tracking(
    candidate_map,
    benchmark_map,
    agreement_map,
    candidate_suffix,
    benchmark_suffix,
    candidate_include,
    candidate_exclude,
    benchmark_include,
    benchmark_exclude,
    expected_df,
    expected_attr,
):
    """Tests attribute tracking functionality"""

    # Test attribute tracking
    results = _attribute_tracking_xarray(
        candidate_map=candidate_map,
        benchmark_map=benchmark_map,
        agreement_map=agreement_map,
        candidate_suffix=candidate_suffix,
        benchmark_suffix=benchmark_suffix,
        candidate_include=candidate_include,
        candidate_exclude=candidate_exclude,
        benchmark_include=benchmark_include,
        benchmark_exclude=benchmark_exclude,
    )

    if agreement_map is None:
        pd.testing.assert_frame_equal(results, expected_df)
    else:
        pd.testing.assert_frame_equal(results[0], expected_df)
        _assert_pairing_dict_equal(results[1].attrs, expected_attr)


@parametrize_with_cases(
    "candidate_map, benchmark_map, candidate_include, candidate_exclude, benchmark_include, benchmark_exclude, exception",
    glob="attribute_tracking_fail",
)
def test_attribute_tracking_fail(
    candidate_map,
    benchmark_map,
    candidate_include,
    candidate_exclude,
    benchmark_include,
    benchmark_exclude,
    exception,
):
    """Tests attribute tracking functionality"""

    # Test attribute tracking
    with raises(exception):
        _attribute_tracking_xarray(
            candidate_map=candidate_map,
            benchmark_map=benchmark_map,
            candidate_include=candidate_include,
            candidate_exclude=candidate_exclude,
            benchmark_include=benchmark_include,
            benchmark_exclude=benchmark_exclude,
        )



"""
Test functionality for compute_probabilistic_metrics.py
"""

__author__ = "Fernando Aristizabal"

from pytest_cases import parametrize_with_cases
from pytest import raises, warns

from tests.conftest import _compare_metrics_df_with_xarray
from gval.comparison.compute_probabilistic_metrics import _compute_probabilistic_metrics


@parametrize_with_cases(
    "candidate_map, benchmark_map, compute_kwargs, expected_df",
    # glob="compute_prob_metrics_*_success",
    glob="compute_prob_metrics_crps_quadrature_success",
)
def test_compute_probabilistic_metrics(
    candidate_map, benchmark_map, compute_kwargs, expected_df
):
    """tests probabilistic metrics computation"""

    # compute categorical metrics
    metrics_df = _compute_probabilistic_metrics(
        candidate_map=candidate_map, benchmark_map=benchmark_map, **compute_kwargs
    )

    _compare_metrics_df_with_xarray(metrics_df, expected_df)


@parametrize_with_cases(
    "candidate_map, benchmark_map, compute_kwargs, expected_error",
    glob="compute_prob_metrics_*_fail",
)
def test_compute_probabilistic_metrics_fail(
    candidate_map, benchmark_map, compute_kwargs, expected_error
):
    """tests probabilistic metrics computation"""

    # compute categorical metrics
    with raises(expected_error):
        _compute_probabilistic_metrics(
            candidate_map=candidate_map, benchmark_map=benchmark_map, **compute_kwargs
        )


@parametrize_with_cases(
    "candidate_map, benchmark_map, compute_kwargs, expected_warning",
    glob="compute_prob_metrics_*_warns",
)
def test_compute_probabilistic_metrics_warns(
    candidate_map, benchmark_map, compute_kwargs, expected_warning
):
    """tests probabilistic metrics computation warnings"""

    # compute categorical metrics
    with warns(expected_warning):
        _compute_probabilistic_metrics(
            candidate_map=candidate_map, benchmark_map=benchmark_map, **compute_kwargs
        )



import numpy as np
from pytest_cases import parametrize_with_cases
from pytest import raises

# added comment because of flake8 believes it's unused
import gval  # noqa: F401


@parametrize_with_cases(
    "candidate_map, crs, entries",
    glob="categorical_plot_success",
)
def test_categorical_plot_success(candidate_map, crs, entries):
    candidate_map.rio.write_crs(crs, inplace=True)
    viz_object = candidate_map.gval.cat_plot(basemap=None)
    assert len(viz_object.axes.get_legend().texts) == entries


@parametrize_with_cases(
    "candidate_map, legend_labels, num_classes",
    glob="categorical_plot_fail",
)
def test_categorical_plot_fail(candidate_map, legend_labels, num_classes):
    candidate_map.data = np.random.choice(np.arange(num_classes), candidate_map.shape)
    with raises(ValueError):
        _ = candidate_map.gval.cat_plot(legend_labels=legend_labels, basemap=None)


@parametrize_with_cases(
    "candidate_map, axes",
    glob="continuous_plot_success",
)
def test_continuous_plot_success(candidate_map, axes):
    viz_object = candidate_map.gval.cont_plot(basemap=None)
    assert len(viz_object.figure.axes) == axes


@parametrize_with_cases(
    "candidate_map",
    glob="continuous_plot_fail",
)
def test_continuous_plot_fail(candidate_map):
    with raises(ValueError):
        _ = candidate_map.gval.cont_plot()



"""
Computes continuous value metrics given an agreement map.
"""

# __all__ = ['*']
__author__ = "Fernando Aristizabal"

from typing import Iterable, Union, List

import numpy as np
import pandera.pandas as pa
import pandas as pd
from pandera.typing import DataFrame
import xarray as xr
import geopandas as gpd
import dask as da

from gval import ContStats
from gval.utils.schemas import Metrics_df, Subsample_identifiers, Sample_identifiers
from gval.utils.loading_datasets import _convert_to_dataset


def _get_masked_data(
    agreement: xr.Dataset,
    candidate: xr.Dataset,
    benchmark: xr.Dataset,
    nodata: list,
    var_name: str,
) -> List[xr.Dataset]:
    """
    Gets masked data for integer valued datasets in order to not process nodata values

    Parameters
    ----------
    agreement : xr.Dataset
        Agreement Map
    candidate : xr.Dataset
        Candidate Map
    benchmark : xr.Dataset
        Benchmark Map
    nodata : list
        Nodata values in the list
    var_name : str
        Name of variable

    Returns
    -------
    List[xr.Dataset, xr.Dataset, xr.Dataset]
        Datasets with selected coordinates
    """

    cmask, bmask = (
        xr.where(candidate[var_name] == nodata, 0, 1),
        xr.where(benchmark[var_name] == nodata, 0, 1),
    )
    tmask = cmask & bmask

    # Select coordinates from xarray
    with da.config.set({"array.slicing.split_large_chunks": True}):
        agreement_sel = agreement[var_name].where(tmask.compute(), drop=True)
        candidate_sel = candidate[var_name].where(tmask.compute(), drop=True)
        benchmark_sel = benchmark[var_name].where(tmask.compute(), drop=True)

    return (
        agreement_sel,
        candidate_sel,
        benchmark_sel,
    )


@pa.check_types
def _compute_continuous_metrics(
    agreement_map: Union[xr.DataArray, xr.Dataset],
    candidate_map: Union[xr.DataArray, xr.Dataset],
    benchmark_map: Union[xr.DataArray, xr.Dataset],
    metrics: Union[str, Iterable[str]] = "all",
    subsampling_average: str = "micro",
    subsampling_df: gpd.GeoDataFrame = None,
    epsilon: float = 1e-10,
) -> DataFrame[Metrics_df]:
    """
    Computes continuous metrics.

    Parameters
    ----------
    agreement_map : Union[xr.DataArray, xr.Dataset, List[Union[xr.DataArray, xr.Dataset]]]
        Agreement map, error based (candidate - benchmark).
    candidate_map : Union[xr.DataArray, xr.Dataset]
        Candidate map.
    benchmark_map : Union[xr.DataArray, xr.Dataset]
        Benchmark map.
    metrics : Union[str, Iterable[str]], default = "all"
        String or list of strings representing metrics to compute.
    subsampling_average : str, default = "micro"
        Strategy to average samples if there is more than one in the agreement map
    subsampling_df : gpd.GeoDataFrame, default = None
        DataFrame with geometries to subsample or use as exclusionary masks and optional sample weights
    epsilon : float, default = 1e-10
        Small value to avoid division by zero in some metrics.

    Returns
    -------
    DataFrame[Metrics_df]
        Metrics DF with computed metrics per sample.

    Raises
    ------
    ValueError
        If metrics is not a string or list of strings.

    References
    ----------
    .. [1] `7th International Verification Methods Workshop <https://www.cawcr.gov.au/projects/verification/>`_
    .. [2] `3.3. Metrics and scoring: quantifying the quality of predictions <https://scikit-learn.org/stable/modules/model_evaluation.html>`_
    """

    if not isinstance(agreement_map, list):
        agreement_map = [agreement_map]
        candidate_map = [candidate_map]
        benchmark_map = [benchmark_map]

    metric_dfs = []
    for idx, (agreement, benchmark, candidate) in enumerate(
        zip(agreement_map, benchmark_map, candidate_map)
    ):
        # Change data to Dataset if DataArray
        agreement = _convert_to_dataset(agreement)
        candidate = _convert_to_dataset(candidate)
        benchmark = _convert_to_dataset(benchmark)

        # Check if integer type and nodata values
        is_int = (
            np.issubdtype(candidate.dtype, np.integer)
            if isinstance(candidate, xr.DataArray)
            else np.issubdtype(candidate["band_1"].dtype, np.integer)
        )
        nodata = [agreement[x].rio.nodata for x in agreement.data_vars]

        # Remove no data value if int type form calculation, otherwise leave all values in
        # Necessary because there is not an int sentinel value
        if is_int and np.all([x is not None for x in nodata]):
            final_stats = []
            # Iterate through each band and gather statistics
            for nodata_idx, var_name in enumerate(agreement.data_vars):
                # Create mask for all nodata values
                agreement_sel, candidate_sel, benchmark_sel = _get_masked_data(
                    agreement, candidate, benchmark, nodata[nodata_idx], var_name
                )

                statistics, names = ContStats.process_statistics(
                    metrics,
                    error=agreement_sel,
                    candidate_map=candidate_sel,
                    benchmark_map=benchmark_sel,
                    epsilon=epsilon,
                )

                del agreement_sel, candidate_sel, benchmark_sel

                final_stats.append(statistics)

            statistics = [
                {f"band_{idx + 1}": val for idx, val in enumerate(lst)}
                for lst in np.array(final_stats).T
            ]

        else:
            statistics, names = ContStats.process_statistics(
                metrics,
                error=agreement,
                candidate_map=candidate,
                benchmark_map=benchmark,
                epsilon=epsilon,
            )

        # create metrics_df
        metric_df = dict()
        for name, stat in zip(names, statistics):
            metric_df[name] = stat

        def is_nested_dict(d):
            # if not isinstance(d, dict):
            #     return False
            return any(isinstance(v, dict) for v in d.values())

        if is_nested_dict(metric_df):
            metric_df = pd.DataFrame.from_dict(metric_df, orient="index").transpose()
            metric_df.reset_index(inplace=True)
            metric_df.rename(columns={"index": "band"}, inplace=True)
            metric_df["band"] = metric_df["band"].str.replace("band_", "")

        else:
            # dataarray
            metric_df = pd.DataFrame(metric_df, index=[0])

            # add band
            metric_df.insert(0, "band", "1")

        if subsampling_df is not None:
            metric_df.insert(0, "subsample", f"{idx + 1}")

        metric_dfs.append(metric_df)

    metric_df = pd.concat(metric_dfs).reset_index().drop(columns=["index"])

    if subsampling_df is not None:
        if subsampling_average == "band":
            metric_df = (
                metric_df.groupby(Subsample_identifiers.columns())
                .mean(numeric_only=True)
                .reset_index()
            )

            metric_df.insert(1, "band", "averaged")

        if subsampling_average == "subsample":
            metric_df = (
                metric_df.groupby(Sample_identifiers.columns())
                .mean(numeric_only=True)
                .reset_index()
            )

            metric_df.insert(0, "subsample", "averaged")

        if subsampling_average == "weighted":
            if subsampling_df.get("weights") is None:
                raise ValueError(
                    "Must have weights if weighted is chosen for subsampling"
                )  # pragma: no cover

            metric_df.loc[:, "weights"] = subsampling_df["weights"]

            # compute weighted average
            weighted_metrics = (
                metric_df.loc[:, metrics]
                .multiply(metric_df.loc[:, "weights"], axis=0)
                .reset_index(drop=True)
            )

            # add weighted metrics to metric_df
            metric_df.loc[:, metrics] = weighted_metrics

            # take average of weighted metrics
            metric_df = (
                metric_df.groupby(Sample_identifiers.columns())
                .sum(numeric_only=True)
                .drop(
                    columns=["weights", "subsample"],
                    errors="ignore",
                )
                .divide(metric_df.loc[:, "weights"].sum())
                .reset_index()
            )

            metric_df.insert(0, "subsample", "averaged")

    return metric_df



import memray
# import gval
import rioxarray as rxr
import subprocess
import os
import time
import rioxarray as rxr
import xarray as xr
import gval
import geopandas as gpd
from gval.homogenize.rasterize import _rasterize_data
from gval.homogenize.vectorize import _vectorize_data
from gval.homogenize.numeric_alignment import _align_numeric_data_type
from typing import Union
from rasterio.enums import Resampling
import time
import logging
from datetime import datetime
import numpy as np
import gc
import pandas as pd
from gval.catalogs.catalogs import catalog_compare


def check_same_type(c_map: Union[xr.Dataset, xr.DataArray], benchmark_map: Union[xr.Dataset, xr.DataArray]):
    """
    Makes sure benchmark map is the same data type as the candidate object

    Parameters
    ----------
    benchmark_map: Union[xr.Dataset, xr.DataArray]
        Benchmark Map

    Raises
    ------
    TypeError

    """
    if not isinstance(benchmark_map,type(c_map)):
        raise TypeError(f"Benchmark Map needs to be data type of {type(c_map)}")


from typing import Optional, Tuple, Union

import xarray as xr
from shapely.geometry import box
from rasterio.enums import Resampling
from odc.geo.xr import ODCExtensionDa

from gval.utils.loading_datasets import _handle_xarray_memory
from gval.utils.exceptions import RasterMisalignment, RastersDontIntersect

ODCExtensionDa


def _matching_crs(
    candidate_map: Union[xr.DataArray, xr.Dataset],
    benchmark_map: Union[xr.DataArray, xr.Dataset],
) -> bool:
    """
    Checks candidate and benchmark maps for matching crs's.

    Parameters
    ----------
    candidate_map : Union[xr.DataArray, xr.Dataset]
        Candidate map.
    benchmark_map : Union[xr.DataArray, xr.Dataset]
        Benchmark map.

    Returns
    -------
    bool
        Matching CRS's or not.
    """
    match = candidate_map.rio.crs == benchmark_map.rio.crs
    return match


def _matching_spatial_indices(
    candidate_map: Union[xr.DataArray, xr.Dataset],
    benchmark_map: Union[xr.DataArray, xr.Dataset],
    raise_exception: bool = False,
) -> bool:
    """
    Checks for matching indices in candidate and benchmark maps.

    TODO: How does this do with 3d DataArrays and Datasets?

    Parameters
    ----------
    candidate_map : Union[xr.DataArray, xr.Dataset]
        Candidate map.
    benchmark_map : Union[xr.DataArray, xr.Dataset]
        Benchmark map.

    Returns
    -------
    bool
        Whether or not maps are perfectly aligned.

    Raises
    ------
    RasterMisalignment
    """
    try:
        xr.align(candidate_map, benchmark_map, join="exact")
    except ValueError:
        aligned = False
    else:
        aligned = True

    # raises exception if not aligned and desired to raise
    if (not aligned) & (raise_exception):
        raise RasterMisalignment

    return aligned


def _rasters_intersect(
    candidate_map: Union[xr.DataArray, xr.Dataset],
    benchmark_map: Union[xr.DataArray, xr.Dataset],
    raise_exception: bool = False,
) -> bool:
    """
    Predicates spatial intersection of two `xr.DataArrays` or `xr.Datasets`.

    CRS's don't have to be the same.

    Parameters
    ----------
    candidate_map : Union[xr.DataArray, xr.Dataset]
        Candidate map.
    benchmark_map : Union[xr.DataArray, xr.Dataset]
        Benchmark map.

    Returns
    -------
    bool
        Tests spatial intersection.

    Raises
    ------
    RasterDontIntersect
        Rasters do not intersect.
    """

    # already matching crs's
    if _matching_crs(candidate_map, benchmark_map):
        candidate_map_bounds = candidate_map.rio.bounds()
        benchmark_map_bounds = benchmark_map.rio.bounds()

    # else transform bounds
    else:
        # use benchmark crs just to test for intersection
        target_crs = benchmark_map.rio.crs

        candidate_map_bounds = candidate_map.rio.transform_bounds(target_crs)
        benchmark_map_bounds = benchmark_map.rio.transform_bounds(target_crs)

    # convert bounds to shapely boxes
    candidate_map_box = box(*candidate_map_bounds)
    benchmark_map_box = box(*benchmark_map_bounds)

    # intersection as bool
    rasters_intersect_bool = candidate_map_box.intersects(benchmark_map_box)

    if (not rasters_intersect_bool) & raise_exception:
        raise RastersDontIntersect
    else:
        return rasters_intersect_bool


def _reproject_map(
    original_map: Union[xr.DataArray, xr.Dataset],
    target_map: Union[xr.DataArray, xr.Dataset],
    resampling: str,
) -> Union[xr.DataArray, xr.Dataset]:
    """

    Parameters
    ----------
    original_map: Union[xr.DataArray, xr.Dataset]
        Map to be reprojected
    target_map: Union[xr.DataArray, xr.Dataset]
        Map to use for extent, resolution, and spatial reference
    resampling: str
        Method to resample changing resolutions

    Returns
    -------
    Union[xr.DataArray, xr.Dataset]
        Reprojected map
    """

    is_dst = isinstance(original_map, xr.Dataset)

    print('ODC reproject', datetime.now().strftime('%M:%S:%f'))
    nodata = target_map["band_1"].rio.nodata if is_dst else target_map.rio.nodata
    reproj = original_map.odc.reproject(
        target_map.odc.geobox, tight=True, dst_nodata=nodata, resampling=resampling
    )
    # reproj = original_map.rio.reproject_match(target_map, resampling=resampling)
    original_map.close()

    print('reproj rename', datetime.now().strftime('%M:%S:%f'))
    # Coordinates need to be aligned
    reproj_coords = (
        reproj.rename({"longitude": "x", "latitude": "y"})
        if "longitude" in reproj.coords
        else reproj
    )

    reproj.close()
    del reproj

    print('assign coords ', datetime.now().strftime('%M:%S:%f'))
    #  Coordinates are virtually the same but 1e-8 or so is rounded differently
    final_reproj = reproj_coords.assign_coords(
        {"x": target_map.coords["x"], "y": target_map.coords["y"]}
    )

    reproj_coords.close()
    del reproj_coords
    return final_reproj


def _align_rasters(
    candidate_map: Union[xr.DataArray, xr.Dataset],
    benchmark_map: Union[xr.DataArray, xr.Dataset],
    target_map: Optional[Union[xr.DataArray, xr.Dataset, str]],
    resampling: Optional[Resampling] = Resampling.nearest,
) -> Tuple[Union[xr.DataArray, xr.Dataset], Union[xr.DataArray, xr.Dataset]]:
    """
    Reprojects raster to match a target map.

    Parameters
    ----------
    candidate_map : Union[xr.DataArray, xr.Dataset]
        Candidate map. Nodata value should be set.
    benchmark_map : Union[xr.DataArray, xr.Dataset]
        Benchmark map. Nodata value should be set.
    target_map : Optional[Union[xr.DataArray, xr.Dataset, str]]
        Target map as string that can have values "candidate" or "benchmark". Otherwise pass an `xr.DataArray` or `xr.Dataset` to align maps to.
    resampling : rasterio.enums.Resampling
        See :func:`rasterio.warp.reproject` for more details.
    **kwargs : dict or keyword arguments
        Dictionary or keyword arguments to be passed `rio.xarray.reproject_match()`. DEPRECATED: This was found to conflict with rxr.rio.reproject_matchrk
        ().
    Returns
    -------
    Tuple[Union[xr.DataArray, xr.Dataset], Union[xr.DataArray, xr.Dataset]]
        Tuple with aligned candidate and benchmark map respectively.

    Raises
    ------
    ValueError
        target_map argument only accepts xr.DataArray, xr.Dataset, 'candidate', 'benchmark'.
    """

    def ensure_nodata_value_is_set(dataset_or_dataarray):
        """
        Ensure nodata values are set.
        NOTE: this prevents arbitrary large numbers from being introduced into the aligned map.
        """

        err_message = (
            "Both candidate and benchmark maps need to have nodata values set."
        )

        # dataarray case
        if isinstance(dataset_or_dataarray, xr.DataArray):
            if dataset_or_dataarray.rio.nodata is None:
                raise ValueError(err_message)

        # dataset case
        elif isinstance(dataset_or_dataarray, xr.Dataset):
            for var_name in dataset_or_dataarray.data_vars.keys():
                if dataset_or_dataarray[var_name].rio.nodata is None:
                    raise ValueError(err_message)
    print('ensure nodata value is set', datetime.now().strftime('%M:%S:%f'))
    # time.sleep(1)
    # ensure both candidate and benchmarks have nodata values set
    ensure_nodata_value_is_set(candidate_map)
    ensure_nodata_value_is_set(benchmark_map)

    print('reprojection', datetime.now().strftime('%M:%S:%f'))
    # time.sleep(1)

    # already matching crs's and indices
    if _matching_crs(candidate_map, benchmark_map) & _matching_spatial_indices(
        candidate_map, benchmark_map
    ):
        return candidate_map, benchmark_map

    # align benchmark and candidate to target
    elif isinstance(target_map, (xr.DataArray, xr.Dataset)):
        candidate_map = _reproject_map(candidate_map, target_map, resampling)
        benchmark_map = _reproject_map(benchmark_map, target_map, resampling)

    # match candidate to benchmark
    elif target_map == "benchmark":
        candidate_map = _reproject_map(candidate_map, benchmark_map, resampling)

    # match benchmark to candidate
    elif target_map == "candidate":
        benchmark_map = _reproject_map(benchmark_map, candidate_map, resampling)

    else:
        raise ValueError(
            "Target_map argument only accepts xr.DataArray or xr.Dataset or a string with values of 'candidate' or 'benchmark'."
        )

    return candidate_map, benchmark_map


def _spatial_alignment(
    candidate_map: Union[xr.DataArray, xr.Dataset],
    benchmark_map: Union[xr.DataArray, xr.Dataset],
    target_map: Optional[Union[xr.DataArray, xr.Dataset, str]] = "benchmark",
    resampling: Optional[Resampling] = Resampling.nearest,
) -> Tuple[Union[xr.DataArray, xr.Dataset], Union[xr.DataArray, xr.Dataset]]:
    """
    Reproject :class:`xarray.Dataset` objects

    .. note:: Only 2D/3D arrays with dimensions 'x'/'y' are currently supported.
        Others are appended as is.
        Requires either a grid mapping variable with 'spatial_ref' or
        a 'crs' attribute to be set containing a valid CRS.
        If using a WKT (e.g. from spatiareference.org), make sure it is an OGC WKT.

    Parameters
    ----------
    candidate_map: Union[xr.DataArray, xr.Dataset]
        Candidate map in xarray DataArray format.
    benchmark_map: Union[xr.DataArray, xr.Dataset]
        Benchmark map in xarray DataArray format.
    target_map: Optional[Union[xr.DataArray, xr.Dataset, str]], default = "benchmark"
        xarray object to match candidates and benchmarks to or str with 'candidate' or 'benchmark' as accepted values.
    resampling : rasterio.enums.Resampling
        See :func:`rasterio.warp.reproject` for more details.
    **kwargs: dict, optional
        Additional keyword arguments to pass into :func:`rasterio.warp.reproject`.
        DEPRECATED: This was found to have issues with spatial alignment if certain arguments were passed.

    Returns
    --------
    Tuple[Union[xr.DataArray, xr.Dataset], Union[xr.DataArray, xr.Dataset]]
        Tuple with candidate and benchmark map respectively.

    Raises
    ------
    RasterDontIntersect
        Rasters do not intersect.

    References
    ----------
    .. [1] `xr.reproject_match <https://corteva.github.io/rioxarray/stable/rioxarray.html#rioxarray.raster_dataset.RasterDataset.reproject>`_
    .. [2] `xr.reproject <https://corteva.github.io/rioxarray/stable/rioxarray.html#rioxarray.raster_dataset.RasterDataset.reproject>`_
    .. [2] `rasterio.warp.reproject <https://rasterio.readthedocs.io/en/latest/api/rasterio.warp.html#rasterio.warp.reproject>`_

    """

    # check if rasters intersect at all
    _rasters_intersect(candidate_map, benchmark_map, raise_exception=True)

    # reproject maps to align
    cam, bem = _align_rasters(candidate_map, benchmark_map, target_map, resampling)

    print('handle xarray memory', datetime.now().strftime('%M:%S:%f'))
    # time.sleep(1)

    return cam, bem
    # tuple(map(lambda x: _handle_xarray_memory(x, make_temp=True), [cam, bem]))


def run():
    TEST_DATA_DIR = '../notebooks/'
    candidate_continuous_catalog = pd.read_csv(f'{TEST_DATA_DIR}candidate_catalog_0.csv')
    benchmark_continuous_catalog = pd.read_csv(f'{TEST_DATA_DIR}benchmark_catalog_0.csv')
    candidate_categorical_catalog = pd.read_csv(f'{TEST_DATA_DIR}candidate_catalog_1.csv')
    benchmark_categorical_catalog = pd.read_csv(f'{TEST_DATA_DIR}benchmark_catalog_1.csv')

    arguments = {
        "candidate_catalog": candidate_categorical_catalog,
        "benchmark_catalog": benchmark_categorical_catalog,
        "on": "compare_id",
        "map_ids": "map_id",
        "how": "inner",
        "compare_type": "categorical",
        "compare_kwargs": {
            "metrics": (
                "critical_success_index",
                "true_positive_rate",
                "positive_predictive_value",
            ),
            "encode_nodata": True,
            "nodata": -9999,
            "positive_categories": 2,
            "negative_categories": 1
        },
        "open_kwargs": {
            "mask_and_scale": True,
            "masked": True
        }
    }

    agreement_categorical_catalog = catalog_compare(**arguments)

    time.sleep(.5)

    # print('close', datetime.now().strftime('%M:%S:%f'))
    # candidate_map.close()
    # benchmark_map.close()


    # time.sleep(.5)
    #
    # print('delete', datetime.now().strftime('%M:%S:%f'))
    # del candidate_map, benchmark_map
    #
    # time.sleep(.5)

    # gc.collect()
    #
    # if isinstance(benchmark_map, gpd.GeoDataFrame):
    #     benchmark_map = _rasterize_data(
    #         candidate_map=candidate_map,
    #         benchmark_map=benchmark_map,
    #         rasterize_attributes=None,
    #     )
    #
    # # agreement_map_format = "vector"
    #
    # check_same_type(candidate_map, benchmark_map)
    #
    # candidate, benchmark = _align_numeric_data_type(
    #     candidate_map=candidate_map, benchmark_map=benchmark_map
    # )
    #
    # candidate, benchmark = _spatial_alignment(
    #     candidate_map=candidate,
    #     benchmark_map=benchmark,
    #     target_map="benchmark",
    #     resampling=Resampling.nearest,
    # )
    #
    # print("dtypes c", candidate_map.dtype, candidate.dtype)
    # print("dtypes b", benchmark_map.dtype, benchmark.dtype)
    #
    # print('agreement map', datetime.now().strftime('%M:%S:%f'))
    # agreement_map = candidate.gval.compute_agreement_map(
    #     benchmark_map=benchmark,
    #     comparison_function='szudzik'
    # )
    #
    # print('close final candidate', datetime.now().strftime('%M:%S:%f'))
    # candidate.rio.to_raster('./cand.tif')
    #
    # del candidate
    #
    # print('close final benchmark', datetime.now().strftime('%M:%S:%f'))
    # benchmark.rio.to_raster('./bench.tif')
    #
    # del benchmark
    #
    # time.sleep(1)
    # print('finish', datetime.now().strftime('%M:%S:%f'))
    #
    # print(agreement_map.dtype)
    #
    # agreement_map.rio.to_raster('./aguint8.tif')

    # ag = rxr.open_rasterio('./ag.tif', mask_and_scale=True)
    # # ag.load()
    #
    # time.sleep(2)

    # cand = rxr.open_rasterio('./c_uint8.tif')
    # print(cand.dtype)
    # cand.load()
    #
    # time.sleep(.5)
    #
    # bench = rxr.open_rasterio('./b_uint8.tif')
    # print(bench.dtype)
    # bench.load()
    #
    # time.sleep(.5)
    #
    # bench2 = rxr.open_rasterio('./aguint8.tif')
    # print(bench2.dtype)
    # bench2.load()
    #
    # time.sleep(.5)


if __name__ == '__main__':

    # import argparse
    # import subprocess
    #
    # # Parse arguments
    # parser = argparse.ArgumentParser(description="Flamegraph baby")
    #
    # parser.add_argument(
    #     "-o", "--file_name", help="REQUIRED: Must be an existing directory", required=True
    # )
    #
    # args = vars(parser.parse_args())
    # file_name = args['file_name']
    #
    file_name = 'one701.bin'

    if os.path.exists(file_name):
        os.remove(file_name)

    with memray.Tracker(file_name):
        # for x in range(5):
        run()

        time.sleep(2)

    subprocess.call(['memray', 'flamegraph', file_name])




import xarray as xr

from gval.accessors.gval_xarray import GVALXarray


@xr.register_dataset_accessor("gval")
class GVALDataset(GVALXarray):
    """
    Class for extending xarray Dataset functionality

    Attributes
    ----------
    _obj : xr.Dataset
       Object to use off the accessor
    data_type : type
        Data type of _obj
    """

    def __init__(self, xarray_obj: xr.Dataset):
        super().__init__(xarray_obj)



"""
Comparison functionality
 - Includes pairing and comparison functions
 - crosstabbing functionality
 - Would np.meshgrid lend itself to pairing problem?
TODO:
    - Have not tested parallel case.
    - How to handle xr.Datasets, multiple bands, and multiple variables.
    - consider making a function registry to store pairing functions and their names in a dictionary
        - [Guide to function registration in Python with decorators](https://blog.miguelgrinberg.com/post/the-ultimate-guide-to-python-decorators-part-i-function-registration)
"""

# __all__ = ['*']
__author__ = "Fernando Aristizabal"


from typing import Iterable, Optional, Union, Tuple, Callable, Dict
from numbers import Number
from itertools import product

import numpy as np
import xarray as xr
import numba as nb
import dask

from gval.comparison.pairing_functions import (
    _make_pairing_dict_fn,
)
from gval.utils.loading_datasets import _handle_xarray_memory, _check_dask_array


def _compute_agreement_map(
    candidate_map: Union[xr.DataArray, xr.Dataset],
    benchmark_map: Union[xr.DataArray, xr.Dataset],
    comparison_function: Union[
        Callable, nb.np.ufunc.dufunc.DUFunc, np.ufunc, np.vectorize, str
    ],
    pairing_dict: Optional[Dict[Tuple[Number, Number], Number]] = None,
    allow_candidate_values: Optional[Iterable[Number]] = None,
    allow_benchmark_values: Optional[Iterable[Number]] = None,
    nodata: Optional[Number] = None,
    encode_nodata: Optional[bool] = False,
    continuous: Optional[bool] = False,
) -> Union[xr.DataArray, xr.Dataset]:
    """
    Computes agreement map as xarray from candidate and benchmark xarray's.

    Parameters
    ----------
    candidate_map : Union[xr.DataArray, xr.Dataset]
        Candidate map.
    benchmark_map : Union[xr.DataArray, xr.Dataset]
        Benchmark map.
    comparison_function : Union[Callable, nb.np.ufunc.dufunc.DUFunc, np.ufunc, np.vectorize, str]
        Comparison function. Created by decorating function with @nb.vectorize() or using np.ufunc(). Use of numba is preferred as it is faster. Strings with registered comparison_functions are also accepted. Possible options include "pairing_dict". If passing "pairing_dict" value, please see the description for the argument for more information on behaviour.
    pairing_dict: Optional[Dict[Tuple[Number, Number], Number]], default = None
        When "pairing_dict" is used for the comparison_function argument, a pairing dictionary can be passed by user. A pairing dictionary is structured as `{(c, b) : a}` where `(c, b)` is a tuple of the candidate and benchmark value pairing, respectively, and `a` is the value for the agreement array to be used for this pairing.

        If None is passed for pairing_dict, the allow_candidate_values and allow_benchmark_values arguments are required. For this case, the pairings in these two iterables will be paired in the order provided and an agreement value will be assigned to each pairing starting with 0 and ending with the number of possible pairings.

        A pairing dictionary can be used by the user to note which values to allow and which to ignore for comparisons. It can also be used to decide how nans are handled for cases where either the candidate and benchmark maps have nans or both.
    allow_candidate_values : Optional[Iterable[Union[int,float]]], default = None
        List of values in candidate to include in computation of agreement map. Remaining values are excluded. If "pairing_dict" is set selected for comparison_function and pairing_function is None, this argument is necessary to construct the dictionary. Otherwise, this argument is optional and by default this value is set to None and all values are considered.
    allow_benchmark_values : Optional[Iterable[Union[int,float]]], default = None
        List of values in benchmark to include in computation of agreement map. Remaining values are excluded. If "pairing_dict" is set selected for comparison_function and pairing_function is None, this argument is necessary to construct the dictionary. Otherwise, this argument is optional and by default this value is set to None and all values are considered.
    nodata : Optional[Number], default = None
        No data value to write to agreement map output. This will use `rxr.rio.write_nodata(nodata)`.
    encode_nodata : Optional[bool], default = False
        Encoded no data value to write to agreement map output. A nodata argument must be passed. This will use `rxr.rio.write_nodata(nodata, encode=encode_nodata)`.

    Returns
    -------
    Union[xr.DataArray, xr.Dataset]
        Agreement map.

    Raises
    ------
    ValueError
        Must pass a value for 'nodata' argument if setting 'encode_nodata' to True.

    References
    ----------
    .. [1] `Creating NumPy universal function <https://numba.readthedocs.io/en/stable/user/vectorize.html>`_
    .. [2] `NumPy vectorize <https://numpy.org/doc/stable/reference/generated/numpy.vectorize.html>`_
    .. [3] `NumPy frompyfunc <https://numpy.org/doc/stable/reference/generated/numpy.frompyfunc.html>`_
    .. [4] `Rioxarray Manage Information Loss <https://corteva.github.io/rioxarray/stable/getting_started/manage_information_loss.html>`_
    .. [5] `Rioxarray NoData Management <https://corteva.github.io/rioxarray/stable/getting_started/nodata_management.html>`_
    """

    """
    TODO:
    - Understand how these arguments affect the agreement map.
    - A partial list of some important arguments is included below with either default or presumed values.
    - These arguments need to be further researched and tested.
    - `keep_attrs` argument should probably be handled here as to be able to preserve both candidate and benchmark attributes. Possibly consider using a prefix or suffix for attribute keys to denote the source map of the attribute (e.g. {candidate_key: value, benchmark_key: value})
    - `dask` needs testing with dask arrays.
    - `dataset_join` needs testing when data variable names in candidate and benchmark differ
    - `output_dtypes` should be considered to manage dtypes for agreement map
    - behavior for agreement map should consider how behavior for attributes, variable names, and dtypes in handled within crosstabbing functionality.
    - read and consider all of the arguments in (xr.apply_ufunc documentation)[https://docs.xarray.dev/en/stable/generated/xarray.apply_ufunc.html]
    - if the attributes or dtypes of the output agreement map are important, the tests must consider these values by the using the correct `xr.testing.assert_*` function.
        - Consider switching test to `xr.testing.assert_identical()` to consider names and attributes.
        - If names are not to be tested, consider using `tests.conftest._assert_pairing_dict_equal()` for attribute testing.
    """

    # some input checking to avoid computing ValueErrors

    # nodata is None & encode_nodata is True
    if nodata is None:
        if encode_nodata:
            raise ValueError(
                "Must pass a value for 'nodata' argument if setting 'encode_nodata' to True."
            )

    # sets kwargs for xr.apply_ufunc
    apply_ufunc_kwargs = {
        "dask": "parallelized",  # how does this work on dask arrays?
        "keep_attrs": True,  # default, copies attrs from first input
        "join": "exact",  # default, raise ValueError instead of aligning when indexes to be aligned are not equal
        "dataset_join": "exact",  # default, data variables on all Dataset objects must match exactly
        "output_dtypes": None,  # default, Optional list of output dtypes. Only used if dask='parallelized' or vectorize=True.
    }

    ############################################################################################

    def _manage_information_loss(agreement_map, crs, nodata, encode_nodata, dtype):
        """
        Manages the information loss due to `xr.apply_ufunc`

        This encapsulated function is to manage information loss that can't be managed with apply_ufunc_kwargs.
        """

        # sets CRS that is lost with `xr.apply_ufunc`
        agreement_map.rio.write_crs(crs, inplace=True)

        # setting agreement map nodata and encoded nodata
        if nodata is not None:
            # this masks the desired nodata value within agreement
            if encode_nodata:
                agreement_map = agreement_map.astype(dtype).where(
                    agreement_map != nodata
                )

                # writes no data and encoded no data if set
                agreement_map.rio.write_nodata(
                    nodata, encoded=encode_nodata, inplace=True
                )
            else:
                agreement_map.rio.write_nodata(nodata, inplace=True)

        return agreement_map

    ############################################################################################
    # Masking out values not in allowed lists

    # mask out values not in allow_candidate_values
    if allow_candidate_values is not None:
        candidate_map = candidate_map.where(candidate_map.isin(allow_candidate_values))

    # mask out values not in allow_benchmark_values
    if allow_benchmark_values is not None:
        benchmark_map = benchmark_map.where(benchmark_map.isin(allow_benchmark_values))

    ###########################################################################################
    # Handling for pairing_dict functionality

    ufunc_args = [candidate_map, benchmark_map]

    if comparison_function == "pairing_dict":  # when pairing_dict is a dict
        # this creates the pairing dictionary from the passed allowed values
        comparison_function = _make_pairing_dict_fn(
            pairing_dict=pairing_dict,
            unique_candidate_values=allow_candidate_values,
            unique_benchmark_values=allow_benchmark_values,
        )

    # use xarray apply ufunc to apply comparison to candidate and benchmark xarray's
    # NOTE: Default behavior loses CRS and uses nodata from first argument (candidate_map)

    agreement_map = xr.apply_ufunc(
        comparison_function, *ufunc_args, **apply_ufunc_kwargs
    )

    is_dask = _check_dask_array(candidate_map)

    def get_unique_values(candidate, benchmark):
        unique_c = (
            dask.array.unique(candidate.data).compute()
            if is_dask
            else np.unique(candidate)
        )
        unique_b = (
            dask.array.unique(benchmark.data).compute()
            if is_dask
            else np.unique(benchmark)
        )

        return unique_c, unique_b

    # Add pairing dictionary and reverse pairing dictionary to agreement map attributes
    if pairing_dict is not None and not continuous:
        agreement_map.attrs["pairing_dictionary"] = pairing_dict

    if pairing_dict is None and not continuous:
        if allow_candidate_values is None or allow_benchmark_values is None:
            if isinstance(candidate_map, xr.Dataset):
                for idx, var in enumerate(candidate_map.data_vars):
                    agreement_map[var].attrs["pairing_dictionary"] = {
                        (x, y): comparison_function(x, y)
                        for x, y in product(
                            *get_unique_values(candidate_map[var], benchmark_map[var])
                        )
                    }

                    if idx == 0:
                        agreement_map.attrs

            else:
                agreement_map.attrs["pairing_dictionary"] = {
                    (x, y): comparison_function(x, y)
                    for x, y in product(
                        *get_unique_values(candidate_map, benchmark_map)
                    )
                }
        else:
            agreement_map.attrs["pairing_dictionary"] = {
                (x, y): comparison_function(x, y)
                for x, y in product(allow_candidate_values, allow_benchmark_values)
            }

    if isinstance(candidate_map, xr.DataArray):
        agreement_map = _manage_information_loss(
            agreement_map=agreement_map,
            crs=candidate_map.rio.crs,
            nodata=nodata,
            encode_nodata=encode_nodata,
            dtype=candidate_map.dtype,
        )
    else:
        for c_var, a_var in zip(candidate_map.data_vars, agreement_map.data_vars):
            agreement_map[a_var] = _manage_information_loss(
                agreement_map=agreement_map[a_var],
                crs=candidate_map[c_var].rio.crs,
                nodata=nodata,
                encode_nodata=encode_nodata,
                dtype=candidate_map[c_var].dtype,
            )

    return _handle_xarray_memory(agreement_map, make_temp=True)



"""
Functions to rasterize vector datasets to process in raster space
"""

from numbers import Number
from typing import Union

import numpy as np
import xarray as xr
import geopandas as gpd
from geocube.api.core import make_geocube

from gval.utils.loading_datasets import _handle_xarray_memory


def _rasterize_data(
    candidate_map: Union[xr.Dataset, xr.DataArray],
    benchmark_map: gpd.GeoDataFrame,
    rasterize_attributes: list,
) -> Union[xr.Dataset, xr.DataArray]:
    """

    Parameters
    ----------
    candidate_map: Union[xr.Dataset, xr.DataArray]
        Candidate map to reference in creation of rasterized benchmark map
    benchmark_map: gpd.GeoDataFrame
        Benchmark map to be rasterized
    rasterize_attributes: list
        Attributes to rasterize

    Returns
    -------
    Union[xr.Dataset, xr.DataArray]
        Rasterized Data

    Raises
    ------
    KeyError

    """

    for attr in rasterize_attributes:
        if (
            not issubclass(type(benchmark_map[attr][0]), Number)
            or rasterize_attributes is None
            or len(rasterize_attributes) == 0
        ):
            raise KeyError(
                "Rasterize attribute needs to be not null and of numeric type"
            )

    rasterized_data = make_geocube(
        vector_data=benchmark_map,
        measurements=rasterize_attributes,
        like=candidate_map,
    )

    # Set nodata to value of candidate map
    if isinstance(candidate_map, xr.DataArray):
        rasterized_data = rasterized_data.to_array()
        # Nodata resolve
        rasterized_data.rio.write_nodata(np.nan, encoded=True, inplace=True)
        rasterized_data.rio.write_nodata(
            candidate_map.rio.encoded_nodata, encoded=True, inplace=True
        )

        if "band" in candidate_map.coords:
            # Deal with tabulation coords check (for single and multi-band examples)
            rasterized_data = rasterized_data.rename({"variable": "band"})
            rast_values = rasterized_data.values
            rasterized_data = rasterized_data.reindex(
                {"band": np.arange(rasterized_data.shape[0]) + 1}
            )
            rasterized_data.values = rast_values

        else:
            rasterized_data = rasterized_data.squeeze().drop_vars("variable")

    else:
        # make sure variable names are the same as would be loaded from rioxarray band_as_variable=True
        rasterized_data = rasterized_data.rename_vars(
            {
                k: v
                for k, v in zip(
                    rasterize_attributes,
                    [
                        f"band_{idx}"
                        for idx in np.arange(len(rasterized_data.data_vars)) + 1
                    ],
                )
            }
        )
        for var_name in rasterized_data.data_vars.keys():
            # Resolve nodata issues

            rasterized_data[var_name] = rasterized_data[var_name].rio.write_nodata(
                np.nan, encoded=True, inplace=True
            )

            if rasterized_data[var_name].rio.encoded_nodata is None:
                rasterized_data[var_name] = rasterized_data[var_name].rio.write_nodata(
                    candidate_map[var_name].rio.encoded_nodata,
                    encoded=True,
                    inplace=True,
                )

    return _handle_xarray_memory(rasterized_data, make_temp=True)



from typing import Union, List, Optional, Tuple

import numpy as np
import dask
import geopandas as gpd
import xarray as xr
from flox.xarray import xarray_reduce
from geocube.api.core import make_geocube


def subsample(
    candidate: Union[xr.DataArray, xr.Dataset],
    benchmark: Union[xr.DataArray, xr.Dataset],
    subsampling_df: Optional[gpd.GeoDataFrame] = None,
) -> Union[
    List[Union[Tuple[xr.DataArray, xr.DataArray], Tuple[xr.Dataset, xr.Dataset]]]
]:
    """
    Method to subsample a candidate and benchmark map

    Parameters
    ----------
    candidate: Union[xr.DataArray, xr.Dataset]
        Candidate map for evaluating
    benchmark: Union[xr.DataArray, xr.Dataset]
        Benchmark map for
    subsampling_df: Optional[gpd.GeoDataFrame], default = None
        Dataframe with columns regarding the geometry and method to subsample

    Raises
    ------
    ValueError
        Sampling_df crs cannot be none

    Returns
    -------
    Union[List[Union[Tuple[xr.DataArray, xr.DataArray], Tuple[xr.Dataset, xr.Dataset]]]]
        Subsampled candidate and benchmark maps or a list of subsampled candidate and benchmark maps

    """

    if subsampling_df.crs is None:
        raise ValueError("sampling_df crs cannot be none")

    sampled_maps = []

    for idx, (_, row) in enumerate(subsampling_df.iterrows()):
        drop, invert = (
            (True, False) if row["subsample_type"] == "include" else (False, True)
        )
        candidate_copy = candidate.rio.clip(
            [row["geometry"]], subsampling_df.crs, drop=drop, invert=invert
        )
        benchmark_copy = benchmark.rio.clip(
            [row["geometry"]], subsampling_df.crs, drop=drop, invert=invert
        )

        if isinstance(candidate, xr.DataArray):
            candidate_copy.attrs["sample_percentage"] = get_subsample_percent(
                subsampling_df.iloc[idx : idx + 1, :], candidate
            )
        else:
            candidate_copy.attrs["sample_percentage"] = [
                get_subsample_percent(
                    subsampling_df.iloc[idx : idx + 1, :],
                    candidate[c_var],
                )
                for c_var in candidate_copy.data_vars
            ]

        # Get sampled maps and percents
        sampled_maps.append([candidate_copy, benchmark_copy])

    return sampled_maps


def get_subsample_percent(
    sampling_df: gpd.GeoDataFrame, og_data: Union[xr.DataArray, xr.Dataset]
) -> float:
    """
    Get percent of original map subsampled by geometry

    Parameters
    ----------
    sampling_df: gpd.GeoDataFrame
        Dataframe with subsample geometries and options
    og_data: Union[xr.DataArray, xr.Dataset]
        Original nodata count

    Returns
    -------
    float
        Percentage of original data the sample covers

    """
    total_size = (
        dask.array.prod(og_data.shape[-2:])
        if og_data.chunks is not None
        else np.prod(og_data.shape[-2:])
    )

    mask = make_geocube(sampling_df, measurements=["subsample_id"], like=og_data)

    try:
        percent = float(
            xarray_reduce(mask["subsample_id"], mask["subsample_id"], func="count")
            / total_size
            * 100
        )
    except IndexError as e:  # pragma: no cover
        raise e("No spatial overlap in subsample geometry and candidate/benchmark data")

    del mask

    return percent



"""
Custom exceptions for gval package
"""

__author__ = "Fernando Aristizabal"


class RasterMisalignment(Exception):  # pragma: no cover
    """Exception raised when rasters don't spatially align."""

    def __init__(self):
        pass

    def __str__(self):
        return "Rasters do not spatially align."


class RastersDontIntersect(Exception):  # pragma: no cover
    """Exception raised when rasters don't spatially intersect."""

    def __init__(self):
        pass

    def __str__(self):
        return "Rasters don't spatially intersect."



"""
Base Statistics Class
"""

from abc import ABC, abstractmethod


class BaseStatistics(ABC):  # pragma: no cover
    """
    Abstract Class for Running Statistics on Agreement Maps

    Attributes
    ----------
    registered_functions : dict
        Available statistical functions for respective statistics class
    signature_validation: dict
        Dictionary with parameter name and types to constrain registered functions
    """

    def __init__(self):
        self.registered_functions = {}
        self.signature_validation = {}

    @abstractmethod
    def register_function(self, name: str):
        """
        Registers function in statistics class

        Parameters
        ----------
        name: str
            Name of function to register in statistics class
        """
        pass

    @abstractmethod
    def available_functions(self) -> list:
        """
        Lists all available functions

        Returns
        -------
        List of available functions
        """
        return self.registered_functions.keys()

    @abstractmethod
    def function_signature_check(self, func):
        """
        Validates signature of registered function

        Returns
        -------
        List of available functions
        """
        pass



# numeric_alignment.py
"""
Functions to check for and ensure numeric datatypes are compatible in numeric processing

"""

from typing import Union, Tuple

import numpy as np
import xarray as xr

from gval.utils.loading_datasets import _handle_xarray_memory


def _align_numeric_dtype(
    candidate_map: xr.DataArray, benchmark_map: xr.DataArray
) -> Tuple[xr.DataArray, xr.DataArray]:
    """
    Align numeric datatypes to the highest precision to avoid warnings and possible issues with numba
    vectorized operations and/or dask operations.

    Parameters
    ----------
    candidate_map: xr.DataArray
        Candidate map xarray DataArray
    benchmark_map: xr.DataArray
        Benchmark map xarray DataArray

    Returns
    -------
    Tuple[xr.DataArray, xr.DataArray]
        Numerically aligned datatype candidate and benchmark maps

    """

    # Determine order for aligning numeric data type
    c_is_int, b_is_int = (
        np.issubdtype(candidate_map.dtype, np.integer),
        np.issubdtype(benchmark_map.dtype, np.integer),
    )
    total_c = candidate_map.dtype.alignment + int(not c_is_int) * 5
    total_b = benchmark_map.dtype.alignment + int(not b_is_int) * 5

    def change_numeric_type(original_map, target_map):
        original_map_data = original_map.data.astype(target_map.data.dtype)
        adjusted_map = original_map.copy(
            data=xr.where(
                original_map == original_map.rio.nodata,
                target_map.data.dtype.type(original_map.rio.nodata),
                original_map_data,
            )
        )

        del original_map_data

        write_nodata = (
            original_map.rio.encoded_nodata
            if original_map.rio.encoded_nodata
            else original_map.rio.nodata
        )
        adjusted_map.rio.write_nodata(
            target_map.data.dtype.type(write_nodata),
            encoded=bool(original_map.rio.encoded_nodata),
            inplace=True,
        )

        return adjusted_map

    # If candidate is higher precision
    if total_c > total_b:
        benchmark_map = change_numeric_type(
            original_map=benchmark_map, target_map=candidate_map
        )
    # If benchmark is higher precision
    if total_b > total_c:
        candidate_map = change_numeric_type(
            original_map=candidate_map, target_map=benchmark_map
        )

    return candidate_map, benchmark_map


def _align_datasets_dtype(
    candidate_map: xr.Dataset, benchmark_map: xr.Dataset
) -> xr.Dataset:
    """
    Iteration through data variables for two datasets to align numerical dtypes

    Parameters
    ----------
    candidate_map: xr.Dataset
        Candidate map xarray dataset
    benchmark_map: xr.Dataset
        Benchmark map xarray dataset

    Returns
    -------
    Tuple[xr.Dataset, xr.Dataset]

    """
    candidate = candidate_map.copy(deep=True)
    benchmark = benchmark_map.copy(deep=True)

    for c_var, b_var in zip(candidate.data_vars, benchmark.data_vars):
        candidate[c_var], benchmark[b_var] = _align_numeric_dtype(
            candidate[c_var], benchmark[b_var]
        )

    return _handle_xarray_memory(candidate, make_temp=True), _handle_xarray_memory(
        benchmark, make_temp=True
    )


def _align_numeric_data_type(
    candidate_map: Union[xr.Dataset, xr.DataArray],
    benchmark_map: Union[xr.Dataset, xr.DataArray],
) -> Tuple[Union[xr.Dataset, xr.DataArray], Union[xr.Dataset, xr.DataArray]]:
    """
    Align data type for xarray objects

    Parameters
    ----------
    candidate_map: Union[xr.Dataset, xr.DataArray]
        Candidate map xarray object
    benchmark_map: Union[xr.Dataset, xr.DataArray]
        Benchmark map xarray dataset

    Returns
    -------
    Tuple[Union[xr.Dataset, xr.DataArray], Union[xr.Dataset, xr.DataArray]]

    """

    if isinstance(candidate_map, xr.DataArray):
        return tuple(
            map(
                lambda x: _handle_xarray_memory(x, make_temp=True),
                _align_numeric_dtype(candidate_map, benchmark_map),
            )
        )
    else:
        return tuple(
            map(
                lambda x: _handle_xarray_memory(x, make_temp=True),
                _align_datasets_dtype(candidate_map, benchmark_map),
            )
        )



from gval.comparison.compute_comparison import ComparisonProcessing
from gval.statistics.categorical_statistics import CategoricalStatistics
from gval.statistics.continuous_statistics import ContinuousStatistics

Comparison = ComparisonProcessing()
CatStats = CategoricalStatistics()
ContStats = ContinuousStatistics()

from gval.accessors import gval_array, gval_dataset, gval_dataframe
from gval.catalogs.catalogs import catalog_compare



"""
Categorical Statistics Functions
"""

import math
from numbers import Number


def true_positive_rate(tp: Number, fn: Number) -> float:
    """
    Computes true positive rate, AKA sensitivity, recall, hit rate

    Parameters
    ----------
    tp: Number
        Count reflecting true positive
    fn: Number
        Count reflecting false negative

    Returns
    -------
    float
        True positive rate from 0 to 1

    References
    ----------
    .. [1] `Sensitivity and Specificity <https://en.wikipedia.org/wiki/Sensitivity_and_specificity>`_
    """
    return tp / (tp + fn)


def true_negative_rate(tn: Number, fp: Number) -> float:
    """
    Computes true negative rate, AKA specificity, selectivity

    Parameters
    ----------
    tn: Number
        Count reflecting true negative
    fp: Number
        Count reflecting false positive

    Returns
    -------
    float
        True negative rate from 0 to 1

    References
    ----------
    .. [1] `Sensitivity and Specificity <https://en.wikipedia.org/wiki/Sensitivity_and_specificity>`_
    """
    return tn / (tn + fp)


def positive_predictive_value(tp: Number, fp: Number) -> float:
    """
    Computes positive predictive value AKA precision

    Parameters
    ----------
    tp: Number
        Count reflecting true positive
    fp: Number
        Count reflecting false positive

    Returns
    -------
    float
        Positive predictive value from 0 to 1

    References
    ----------
    .. [1] `Sensitivity and Specificity <https://en.wikipedia.org/wiki/Sensitivity_and_specificity>`_
    """
    return tp / (tp + fp)


def negative_predictive_value(tn: Number, fn: Number) -> float:
    """
    Computes negative predictive value

    Parameters
    ----------
    tn: Number
        Count reflecting true negative
    fn: Number
        Count reflecting false negative

    Returns
    -------
    float
        Negative predictive value from 0 to 1

    References
    ----------
    .. [1] `Sensitivity and Specificity <https://en.wikipedia.org/wiki/Sensitivity_and_specificity>`_
    """
    return tn / (tn + fn)


def false_negative_rate(tp: Number, fn: Number) -> float:
    """
    Computes false negative rate

    Parameters
    ----------
    tp: Number
        Count reflecting true positive
    fn: Number
        Count reflecting false negative

    Returns
    -------
    float
        False negative rate from 0 to 1

    References
    ----------
    .. [1] `Type I and Type II Error <https://en.wikipedia.org/wiki/Type_I_and_type_II_errors#False_positive_and_false_negative_rates>`_
    """
    return fn / (fn + tp)


def false_positive_rate(tn: Number, fp: Number) -> float:
    """
    Computes false positive rate AKA fall-out

    Parameters
    ----------
    tn: Number
        Count reflecting true negative
    fp: Number
        Count reflecting false positive

    Returns
    -------
    float
        False positive rate from 0 to 1

    References
    ----------
    .. [1] `Type I and Type II Errors <https://en.wikipedia.org/wiki/Type_I_and_type_II_errors#False_positive_and_false_negative_rates>`_
    """
    return fp / (fp + tn)


def false_discovery_rate(tp: Number, fp: Number) -> float:
    """
    Computes false discovery rate

    Parameters
    ----------
    tp: Number
        Count reflecting true positive
    fp: Number
        Count reflecting false positive

    Returns
    -------
    float
        False discovery rate from 0 to 1

    References
    ----------
    .. [1] `False Discovery Rate <https://en.wikipedia.org/wiki/False_discovery_rate>`_
    """
    return fp / (fp + tp)


def false_omission_rate(tn: Number, fn: Number) -> float:
    """
    Computes false omission rate

    Parameters
    ----------
    tn: Number
        Count reflecting true negative
    fn: Number
        Count reflecting false negative

    Returns
    -------
    float
        False omission rate from 0 to 1

    References
    ----------
    .. [1] `Positive and Negative Predictive Values <https://en.wikipedia.org/wiki/Positive_and_negative_predictive_values>`_
    """
    return fn / (fn + tn)


def positive_likelihood_ratio(tp: Number, tn: Number, fp: Number, fn: Number) -> float:
    """
    Computes positive likelihood ratio

    Parameters
    ----------
    tp: Number
        Count reflecting true positive
    tn: Number
        Count reflecting true negative
    fp: Number
        Count reflecting false positive
    fn: Number
        Count reflecting false negative

    Returns
    -------
    float
        Positive likelihood rate from 1 to infinity

    References
    ----------
    .. [1] `Likelihood Ratios <https://en.wikipedia.org/wiki/Likelihood_ratios_in_diagnostic_testing#positive_likelihood_ratio>`_
    """
    return (tp / (tp + fn)) / (fp / (fp + tn))


def negative_likelihood_ratio(tp: Number, tn: Number, fp: Number, fn: Number) -> float:
    """
    Computes negative likelihood ratio

    Parameters
    ----------
    tp: Number
        Count reflecting true positive
    tn: Number
        Count reflecting true negative
    fp: Number
        Count reflecting false positive
    fn: Number
        Count reflecting false negative

    Returns
    -------
    float
        Negative likelihood from 1 to infinity

    References
    ----------
    .. [1] `Likelihood Ratios <https://en.wikipedia.org/wiki/Likelihood_ratios_in_diagnostic_testing#positive_likelihood_ratio>`_
    """
    return (fn / (fn + tp)) / (tn / (tn + fp))


def prevalence_threshold(tp: Number, tn: Number, fp: Number, fn: Number) -> float:
    """
    Computes prevalence threshold

    Parameters
    ----------
    tp: Number
        Count reflecting true positive
    tn: Number
        Count reflecting true negative
    fp: Number
        Count reflecting false positive
    fn: Number
        Count reflecting false negative

    Returns
    -------
    float
        Prevalence threshold from 0 to 1

    References
    ----------
    .. [1] `Prevalence Threshold <https://www.ncbi.nlm.nih.gov/pmc/articles/PMC7540853/>`_
    """
    return math.sqrt(fp / (fp + tn)) / (
        math.sqrt(tp / (tp + fn)) + math.sqrt(fp / (fp + tn))
    )


def critical_success_index(tp: Number, fp: Number, fn: Number) -> float:
    """
    Computes critical success index

    https://www.weather.gov/media/erh/ta2004-03.pdf

    Parameters
    ----------
    tp: Number
        Count reflecting true positive
    fp: Number
        Count reflecting false positive
    fn: Number
        Count reflecting false negative

    Returns
    -------
    float
        Critical success index from 0 to 1

    References
    ----------
    .. [1] `Critical Success Index <https://www.swpc.noaa.gov/sites/default/files/images/u30/Forecast%20Verification%20Glossary.pdf#page=4>`_
    """
    return tp / (tp + fn + fp)


def prevalence(tp: Number, tn: Number, fp: Number, fn: Number) -> float:
    """
    Computes prevalence

    Parameters
    ----------
    tp: Number
        Count reflecting true positive
    tn: Number
        Count reflecting true negative
    fp: Number
        Count reflecting false positive
    fn: Number
        Count reflecting false negative

    Returns
    -------
    float
        Prevalence from 0 to 1

    References
    ----------
    .. [1] `Prevalence <https://en.wikipedia.org/wiki/Prevalence>`_
    """
    return (tp + fn) / (tp + fp + tn + fn)


def accuracy(tp: Number, tn: Number, fp: Number, fn: Number) -> float:
    """
    Computes accuracy

    Parameters
    ----------
    tp: Number
        Count reflecting true positive
    tn: Number
        Count reflecting true negative
    fp: Number
        Count reflecting false positive
    fn: Number
        Count reflecting false negative

    Returns
    -------
    float
        Accuracy from 0 to 1

    References
    ----------
    .. [1] `Accuracy and Precision <https://en.wikipedia.org/wiki/Accuracy_and_precision>`_
    """
    return (tp + tn) / (tp + fp + tn + fn)


def f_score(tp: Number, fp: Number, fn: Number) -> float:
    """
    Computes F-score AKA harmonic mean of precision and sensitivity

    Parameters
    ----------
    tp: Number
        Count reflecting true positive
    fp: Number
        Count reflecting false positive
    fn: Number
        Count reflecting false negative

    Returns
    -------
    float
        F-score from 0 to 1

    References
    ----------
    .. [1] `F-score <https://en.wikipedia.org/wiki/F-score>`_
    """
    return 2 * tp / (2 * tp + fp + fn)


def matthews_correlation_coefficient(
    tp: Number, tn: Number, fp: Number, fn: Number
) -> float:
    """
    Computes matthews correlation coefficient, accounting for accuracy and
    precision of both true positives and true negatives AKA Phi Coefficient

    Parameters
    ----------
    tp: Number
        Count reflecting true positive
    tn: Number
        Count reflecting true negative
    fp: Number
        Count reflecting false positive
    fn: Number
        Count reflecting false negative

    Returns
    -------
    float
        Correlation coefficient from -1 to 1

    References
    ----------
    .. [1] `Matthews Correlation Coefficient <https://en.wikipedia.org/wiki/Phi_coefficient>`_
    """
    return (tp * tn - fp * fn) / math.sqrt(
        (tp + fp) * (tp + fn) * (tn + fp) * (tn + fn)
    )


def fowlkes_mallows_index(tp: Number, fp: Number, fn: Number) -> float:
    """
    Computes Fowlkes-Mallows index

    Parameters
    ----------
    tp: Number
        Count reflecting true positive
    fp: Number
        Count reflecting false positive
    fn: Number
        Count reflecting false negative

    Returns
    -------
    float
        Correlation coefficient from -1 to 1

    References
    ----------
    .. [1] `Fowlkes-Mallows Index <https://en.wikipedia.org/wiki/Fowlkes%E2%80%93Mallows_index>`_
    """
    return math.sqrt((tp / (tp + fp)) * (tp / (tp + fn)))


def equitable_threat_score(tp: Number, tn: Number, fp: Number, fn: Number) -> float:
    """
    Computes Equitable Threat Score (Gilbert Score)

    Parameters
    ----------
    tp: Number
        Count reflecting true positive
    tn: Number
        Count reflecting true negative
    fp: Number
        Count reflecting false positive
    fn: Number
        Count reflecting false negative

    Returns
    -------
    float
        Equitable threat score from -1/3 to 1

    References
    ----------
    .. [1] `Equitable Threat Score <https://resources.eumetrain.org/data/4/451/english/msg/ver_categ_forec/uos2/uos2_ko4.htm>`_

    """
    total_population = tp + tn + fp + fn
    a_ref = ((tp + fp) * (tp + fn)) / total_population
    return (tp - a_ref) / (tp - a_ref + fp + fn)


def balanced_accuracy(tp: Number, tn: Number, fp: Number, fn: Number) -> float:
    """
    Computes Balanced Accuracy

    Parameters
    ----------
    tp: Number
        Count reflecting true positive
    tn: Number
        Count reflecting true negative
    fp: Number
        Count reflecting false positive
    fn: Number
        Count reflecting false negative

    Returns
    -------
    float
        Balanced Accuracy from 0 to 1

    References
    ----------
    .. [1] `Balanced Accuracy <https://neptune.ai/blog/balanced-accuracy#Balanced%20Accuracy>`_
    """

    return ((tp / (tp + fn)) + (tn / (tn + fp))) / 2


def overall_bias(tp: Number, fp: Number, fn: Number) -> float:
    """
    Computes the degree of correspondence between the mean forecast and the mean observation.

    Parameters
    ----------
    tp: Number
        Count reflecting true positive
    fp: Number
        Count reflecting false positive
    fn: Number
        Count reflecting false negative

    Returns
    -------
    float
        Overall Bias

    References
    ----------
    .. [1] `Forecast Verification Glossary <https://www.swpc.noaa.gov/sites/default/files/images/u30/Forecast%20Verification%20Glossary.pdf>`_

    """

    return (tp + fp) / (tp + fn)



from numbers import Number
from typing import Union, Iterable, Optional, List, Dict

import pandas as pd
from shapely import Geometry
import geopandas as gpd
from pandera.typing import DataFrame
import xarray as xr

from gval.comparison.compute_categorical_metrics import _compute_categorical_metrics
from gval.utils.schemas import Metrics_df, SubsamplingDf
from gval.homogenize.rasterize import _rasterize_data


@pd.api.extensions.register_dataframe_accessor("gval")
class GVALDataFrame:
    """
    Class for extending pandas DataFrame functionality

    Attributes
    ----------
    _obj : pd.DataFrame
        Object to use off the accessor
    """

    def __init__(self, pandas_obj):
        self._obj = pandas_obj

    def compute_categorical_metrics(
        self,
        positive_categories: Union[
            Number, Iterable[Number], Dict[str, Union[Number, Iterable[Number]]]
        ],
        negative_categories: Union[
            Number, Iterable[Number], Dict[str, Union[Number, Iterable[Number]]]
        ],
        metrics: Union[str, Iterable[str]] = "all",
        average: str = "micro",
        weights: Optional[Iterable[Number]] = None,
        subsampling_average: Optional[str] = None,
    ) -> DataFrame[Metrics_df]:
        """
        Computes categorical metrics from a crosstab df.

        Parameters
        ----------
        crosstab_df : DataFrame[Crosstab_df]
            Crosstab DataFrame with candidate, benchmark, and agreement values as well as the counts for each occurrence.
        positive_categories : Optional[Union[Number, Iterable[Number], Dict[str, Union[Number, Iterable[Number]]]]]
            Number or list of numbers representing the values to consider as the positive condition. For average types "macro" and "weighted", this represents the categories to compute metrics for.
        negative_categories : Optional[UUnion[Number, Iterable[Number], Dict[str, Union[Number, Iterable[Number]]]]], default = None
            Number or list of numbers representing the values to consider as the negative condition. This should be set to None when no negative categories are used or when the average type is "macro" or "weighted".
        metrics : Union[str, Iterable[str]], default = "all"
            String or list of strings representing metrics to compute.
        average : str, default = "micro"
            Type of average to use when computing metrics. Options are "micro", "macro", and "weighted".
            Micro weighing computes the conditions, tp, tn, fp, and fn, for each category and then sums them.
            Macro weighing computes the metrics for each category then averages them.
            Weighted average computes the metrics for each category then averages them weighted by the number of weights argument in each category.
        weights : Optional[Iterable[Number]], default = None
            Weights to use when computing weighted average. Elements correspond to positive categories in order.

            Example:

            `positive_categories = [1, 2]; weights = [0.25, 0.75]`
        subsampling_average: Optional[str], default = None
            Way to aggregate statistics for subsamples if provided. Options are "sample", "band", and "full-detail"
            Sample calculates metrics and averages the results by subsample
            Band calculates metrics and averages all the metrics by band
            Full-detail does not aggregation on subsample or band


        Returns
        -------
        DataFrame[Metrics_df]
            Metrics DF with computed metrics per sample.

        Raises
        ------
        ValueError
            Value is shared in positive and negative categories.
        ValueError
            Category not found in crosstab df.
        ValueError
            Cannot use average type with only one positive category.
        ValueError
            Number of weights must be the same as the number of positive categories.
        ValueError
            Cannot use average type with negative_categories as not None. Set negative_categories to None for this average type.

        References
        ----------
        .. [1] `Evaluation of binary classifiers <https://en.wikipedia.org/wiki/Evaluation_of_binary_classifiers>`_
        .. [2] `7th International Verification Methods Workshop <https://www.cawcr.gov.au/projects/verification/#Contingency_table>`_
        .. [3] `3.3. Metrics and scoring: quantifying the quality of predictions <https://scikit-learn.org/stable/modules/model_evaluation.html>`_
        """

        return _compute_categorical_metrics(
            crosstab_df=self._obj,
            metrics=metrics,
            positive_categories=positive_categories,
            negative_categories=negative_categories,
            average=average,
            weights=weights,
            sampling_average=subsampling_average,
        )

    def rasterize_data(
        self, reference_map: Union[xr.Dataset, xr.DataArray], rasterize_attributes: list
    ) -> Union[xr.Dataset, xr.DataArray]:
        """
        Convenience function for rasterizing vector data using a reference raster.  For more control use `make_geocube`
        from the geocube package.

        Parameters
        ----------
        reference_map: Union[xr.Dataset, xr.DataArray]
            Map to reference in creation of rasterized vector map
        rasterize_attributes: list
            Attributes to rasterize

        Returns
        -------
        Union[xr.Dataset, xr.DataArray]
            Rasterized Data

        Raises
        ------
        KeyError

        References
        ----------
        .. [1] `make_geocube <https://corteva.github.io/geocube/html/geocube.html>`_

        """

        return _rasterize_data(
            candidate_map=reference_map,
            benchmark_map=self._obj,
            rasterize_attributes=rasterize_attributes,
        )

    def create_subsampling_df(
        self,
        geometries: List[Geometry] = None,
        crs: str = None,
        subsampling_type: Union[str, List[str]] = "exclude",
        subsampling_weights: List[Union[int, float]] = None,
        inplace: bool = False,
    ) -> Union[None, SubsamplingDf]:
        """
        Parameters
        __________
        geometries: List[Geometry], default = None
            Geometries if none are already in the GeoDataFrame
        crs: str
            The spatial reference for the geometries provided
        subsampling_type: Union[str, List[str]], default = "exclude"
            Whether each geometry should be an inclusive subsample or an exclusionary mask
        subsampling_weights: List[Union[int, float]], default = None
            Values to scale the numeric impact of a particular sample
        inplace: bool, default = False
            Whether to adjust the GeoDataFrame calling the operation or a return a new one

        Raises
        ------
        ValueError
            List provided has more or less entries than the DataFrame
        TypeError
            CRS must be provided if geometries are provided

        Returns
        -------
        Union[None, SubsamplingDf]
            GeoDataFrame adhering to subsampling dataframe if not inplace, otherwise None

        """

        geo_df = self._obj if inplace else gpd.GeoDataFrame()
        crs = crs if crs is not None else self._obj.crs

        if (geometries and inplace) or ("geometry" in self._obj.columns):
            if crs is None:
                raise TypeError(
                    "Either provide CRS for or give the original DataFrame a crs if inplace is True"
                )  # pragma: no cover

            if "geometry" not in geo_df.columns:
                geo_df.set_geometry(
                    geometries if geometries is not None else self._obj["geometry"],
                    inplace=True,
                )
                geo_df.crs = crs

        if subsampling_type:
            geo_df["subsample_type"] = subsampling_type

        if subsampling_weights:
            geo_df["weights"] = subsampling_weights

        geo_df.loc[:, "subsample_id"] = geo_df.index.values + 1

        if not inplace:
            return geo_df



"""
Continuous Statistics Functions From Error Based Agreement Maps.
"""

from typing import Callable, Union, Optional
from functools import wraps

import xarray as xr

from gval.utils.loading_datasets import _check_dask_array


def convert_output(func: Callable) -> Callable:  # pragma: no cover
    """
    Decorator that converts the output of a function to a different type.

    If the output is an xarray.DataArray, it is converted to a single numeric value.
    If the output is an xarray.Dataset, it is converted to a dictionary with the band names as keys.

    Parameters
    ----------
    func : Callable
        The function whose output is to be converted.

    Returns
    -------
    wrapper : Callable
        The decorated function.
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        # Call the decorated function
        result = func(*args, **kwargs)

        is_dsk = _check_dask_array(result)
        if isinstance(result, xr.DataArray):
            # Convert to a single numeric value
            return result.compute().item() if is_dsk else result.item()
        elif isinstance(result, xr.Dataset):
            # Convert to a dictionary with band names as keys
            if is_dsk:
                return {
                    band: result[band].compute().item() for band in result.data_vars
                }
            else:
                return {band: result[band].item() for band in result.data_vars}

        return result

    return wrapper


def compute_error_if_none(func: Callable) -> Callable:  # pragma: no cover
    """
    Decorator to compute the error as the difference between the candidate map and benchmark map
    if the error is not supplied.

    The decorated function should have three positional arguments:
    error: Union[xr.DataArray, xr.Dataset], optional
        Candidate minus benchmark error.
    benchmark_map: Union[xr.DataArray, xr.Dataset], optional
        Benchmark map.
    candidate_map: Union[xr.DataArray, xr.Dataset], optional
        Candidate map.

    The error is computed as the difference between the candidate_map and the benchmark_map.
    If the error is provided, it is used directly.

    If none of the arguments are provided, a ValueError is raised.

    Parameters
    ----------
    func : Callable
        Function to be decorated.

    Returns
    -------
    wrapper : Callable
        The decorated function.
    """

    def wrapper(
        error: Optional[Union[xr.DataArray, xr.Dataset]] = None,
        benchmark_map: Optional[Union[xr.DataArray, xr.Dataset]] = None,
        candidate_map: Optional[Union[xr.DataArray, xr.Dataset]] = None,
    ):
        if error is None:
            if benchmark_map is not None and candidate_map is not None:
                error = candidate_map - benchmark_map
            else:
                raise ValueError(
                    "Must provide either `error` or both `candidate_map` and `benchmark_map`."
                )
        return func(error, benchmark_map, candidate_map)

    return wrapper
